#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <zlib.h>
#include <math.h>
#include <iomanip>
#include <list>
#include <vector>
#include <errno.h>
#include <pthread.h>

#include <Inventor/SoInput.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoResetTransform.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/nodes/SoQuadMesh.h>
#include <Inventor/nodes/SoTextureCoordinate2.h>
#include <Inventor/nodes/SoTextureCoordinateBinding.h>
#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/nodes/SoCube.h>

#define MAX2(a,b) ( ((a)>(b)) ? (a) : (b) )
#define MIN2(a,b) ( ((a)<(b)) ? (a) : (b) )
#define TIME(x) { struct timeval start, end;  \
                gettimeofday(&start, NULL); \
                x;                          \
 	        gettimeofday(&end, NULL);   \
	        cerr << "Time: "  << delay(start, end) << "ms" << endl; }

using namespace std;

static char version[80] =
"$Revision: 1.18 $";

// variables
bool animPause = FALSE;
SoSeparator *scene = NULL;
char *textureFilename = "texture.jpg";
char *outFile = "out.iv";
float visc = 0.001;
float dt = 0.01;
float t = 0.0;
int dim[3] = { 20, 20, 20 };
int numParticles = 0;
int desiredNumParticles = 100;
SbVec3f origin(0.0, 0.0, 0.0);
SbVec3f length(1.0, 1.0, 1.0);
SbVec3f voxel    = SbVec3f(length[0]/dim[0], length[1]/dim[1], length[2]/dim[2]);
SbVec3f invvoxel = SbVec3f(dim[0]/length[0], dim[1]/length[1], dim[2]/length[2]);
float *u, *u0, *v, *v0, *w, *w0, *dens, *dens0, *temp, *temp0;
float tempFloat = 0.0;
int tempInt = 0, tempInt2 = 0;
float vortEps = 10.0;
float bouyMult = 0.15; // m/Ks^2, from Nguyen at al.
SbVec3f *vortg;  // grid and particle based vorticities
vector<SbVec3f> vortp;
SbVec3f *du, *dv, *dw;        // velocity derivatives
SoTextureCoordinate2 *texCoord, *texCoord0;
vector<SoSeparator *> partSeps;
SoSeparator *partDebug = NULL;
float maxSpeed = 0.0;
const float maxTemp = 4000; // K
SoMaterial **planeMat = NULL;
SoTexture2 **planeTexs = NULL;
SoSwitch *texSwitch = NULL;
SoSwitch *partSwitch = NULL;
enum {SHOW_FORCE, SHOW_SPEED, SHOW_DENSITY, SHOW_TEXTURE};
int showItem = SHOW_DENSITY;
int showParticles = 0;
int useVorParticles = 0;
const int showItems = 4;
const float ambientTemp = 30.0+273.0; // K
int fileNo=0;
int writeDump = 1;
char *dumpName = "none";
string cmdLine = "";
int renderQuit = 0;
int numThreads = 1;
float maxTime = 1e30;
float maxDensity = 1e-30;
float heatExponent = 0.0;

void addSmoke();
void addPuff(SbVec3f x, float r);
void resetSim();
void dump(float *x, char *name, int t);
void showRay(SbVec3f pos, SbVec3f dir,
	     double height=1, SbColor color=SbColor(-1,-1,-1), int clear=0);
void addParticles(int addNum);


// Compute the delay between t1 and t2 in milliseconds
inline long delay (struct timeval t1, struct timeval t2){
  long d;
    
  d = (t2.tv_sec - t1.tv_sec) * 1000;
  d+= ((t2.tv_usec - t1.tv_usec + 500) / 1000);
  return(d);
}

// clamp definition
// restricts a number to a range [min, max]
template <class T> inline
T clamp(T x, T min, T max){
  if (x<min)
    return(min);
  else if (x>max)
    return(max);
  else
    return(x);
}

inline float finite(SbVec3f &f){
  return (finite(f[0]) && finite(f[1]) && finite(f[2]));
}

inline float gaussFilt(SbVec3f &x, SbVec3f &xp, float rad){
  float r2 = (xp-x).dot(xp-x);
  float sig = rad; //0.5f*rad;
  if (r2<=rad*rad)
    return exp(-r2/(2.f*sig*sig)) / (sig*sig*sig*powf(2.f*M_PI,1.5f));
  else
    return(0.0f);
}

inline float clampin(const float x, const float a, const float b){

  // sort
  if (a<b){
    if (x<a)
      return(a);
    else if (x>b)
      return(b);
  }
  else {
    if (x<b)
      return(b);
    else if (x>a)
      return(a);
  }

  return(x);

}

#define floor(x) ((x)>=0.0?((int)(x)):(-((int)(1-(x)))))

// interpolates f at a point t (in [0,1]) between f1 and f2
float interp(const float& f0, const float& f1, const float& f2, const float& f3, float t){

  // linear interp
  //return ((1-t)*f1 + t*f2);
		    
  // need dk, dkp1, delk
  float dk, dkp1, delk;
  delk = f2-f1;

  if (delk == 0.0f)
    dk = dkp1 = 0.0f;
  else {

    // k = 1
    dk = (f2-f0)*0.5f;
    if (dk * delk <= 0.0f)
      dk = 0.0;

    // k = 2
    dkp1 = (f3-f1)*0.5f;
    if (dkp1 * delk <= 0.0f)
      dkp1 = 0.0f;
  }

  float a0, a1, a2, a3;

  a3 = dk + dkp1 - 2.0f*delk;
  a2 = 3.0f*delk - 2.0f*dk - dkp1;
  a1 = dk;
  a0 = f1;

  float val = (((a3)*t+a2)*t+a1)*t+a0;
  return (clampin(val,f1,f2));
  
}


int *Fx, *Fy, *Fz;
int precomputeIdxs(){

  int Nx=dim[0]+2, Ny=dim[1]+2, Nz=dim[2]+2;
  const int m = 2;
  const int n = 5;
  const int mn = m*n;
  int Xmn=Nx/mn, Ymn=Ny/mn, Zmn=Nz/mn;
  if (Xmn*mn<Nx)
    Xmn++;
  if (Ymn*mn<Ny)
    Ymn++;
  if (Zmn*mn<Nz)
    Zmn++;
  Nx = mn*Xmn;
  Ny = mn*Ymn;
  Nz = mn*Zmn;

  cerr << "Nx=" << Nx
       << " Ny=" << Ny
       << " Nz=" << Nz << " ";
  
  // precompute index tables
  Fx = new int[Nx];
  Fy = new int[Ny];
  Fz = new int[Nz];

  for (int x=0; x<Nx; x++)
    Fx[x] =
      ((x/n)/m)* n*n*n* m*m*m* ((Nz/n)/m)* ((Ny/n)/m) +
      ((x/n)%m)* n*n*n* m*m +
      (x%n)* n*n;
      
  for (int y=0; y<Ny; y++)
    Fy[y] =
      ((y/n)/m)* n*n*n* m*m*m* ((Nz/n)/m) +
      ((y/n)%m)* n*n*n* m +
      (y%n)* n;

  for (int z=0; z<Nz; z++)
    Fz[z] =
      ((z/n)/m)* n*n*n* m*m*m +
      ((z/n)%m)* n*n*n +
      (z%n);

  return(Nx*Ny*Nz);
}

// inline int ix(int i, int j, int k){
//   return(Fx[i] + Fy[j] + Fz[k]);
// }

inline int ix(int i, int j, int k){
  return (i+j*(dim[0]+2)+k*(dim[1]+2)*(dim[0]+2));
}

inline void idx2pos(int i, int j, int k, SbVec3f &x){
  x[0] = (i-0.5f)*voxel[0];
  x[1] = (j-0.5f)*voxel[1];
  x[2] = (k-0.5f)*voxel[2];
}

inline void pos2idx(SbVec3f x, int &i, int &j, int &k){
  for (int m=0; m<3; m++){
    //if (x[m]<origin[m]) x[m]=origin[m]; else if (x[m]>length[m]) x[m] = length[m];
    x[m] = x[m]*invvoxel[m]+0.5f;
  }

  i = floor(x[0]);
  j = floor(x[1]);
  k = floor(x[2]);
}

inline SbVec3f idx2pos(int &i, int &j, int &k){
  SbVec3f x;
  x[0] = (i-0.5f)*voxel[0];
  x[1] = (j-0.5f)*voxel[1];
  x[2] = (k-0.5f)*voxel[2];
  return(x);
}


inline void pos2idx(float x0, float y0, float z0,
		    int &i0, int &j0, int &k0,
		    int &i1, int &j1, int &k1,
		    float &s, float &t, float &q){

  // before: 0.0-1.0, 0...N-1  y=(i+0.5)*1/N  i=N*y-0.5
  // after:  0.0-1.0, 1...N    y=(i-0.5)*1/N  i=N*y+0.5
  // x  -1.0 0.0 1.0
  //      |___|___|___|    x [0.0, 1.0]
  // i      0   1   2      i [1, N]
  // i=1,x=.5h x=0,i=.5  

  if (x0<origin[0]) x0=origin[0]; else if (x0>length[0]) x0 = length[0];
  if (y0<origin[1]) y0=origin[1]; else if (y0>length[1]) y0 = length[1];
  if (z0<origin[2]) z0=origin[2]; else if (z0>length[2]) z0 = length[2];
  
  x0 = x0*invvoxel[0]+0.5f;
  y0 = y0*invvoxel[1]+0.5f;
  z0 = z0*invvoxel[2]+0.5f;

  // voxel index and fraction
  i0 = floor(x0);
  s  = x0-i0;
  i1 = i0+1;

  j0 = floor(y0);
  t  = y0-j0;
  j1 = j0+1;

  k0 = floor(z0);
  q  = z0-k0;
  k1 = k0+1;
  
}


// trilinear interpolation
template <class T>
T getValLinear(float x, float y, float z, const T *u){

  int i[4], j[4], k[4];
  float fj[4], fk[4], s[3];
  
  pos2idx(x, y, z, i[1], j[1], k[1], i[2], j[2], k[2], s[0], s[1], s[2]);

  return(
    (1.0f-s[0])*(1.0f-s[1])*(1.0f-s[2])*u[ix(i[1],j[1],k[1])] +
    (     s[0])*(1.0f-s[1])*(1.0f-s[2])*u[ix(i[2],j[1],k[1])] +
    (1.0f-s[0])*(     s[1])*(1.0f-s[2])*u[ix(i[1],j[2],k[1])] +
    (     s[0])*(     s[1])*(1.0f-s[2])*u[ix(i[2],j[2],k[1])] +
    (1.0f-s[0])*(1.0f-s[1])*(     s[2])*u[ix(i[1],j[1],k[2])] +
    (     s[0])*(1.0f-s[1])*(     s[2])*u[ix(i[2],j[1],k[2])] +
    (1.0f-s[0])*(     s[1])*(     s[2])*u[ix(i[1],j[2],k[2])] +
    (     s[0])*(     s[1])*(     s[2])*u[ix(i[2],j[2],k[2])] );
}


float getVal(float x, float y, float z, const float *u){

  int i[4], j[4], k[4];
  float fj[4], fk[4], s[3];
  
  pos2idx(x, y, z, i[1], j[1], k[1], i[2], j[2], k[2], s[0], s[1], s[2]);

  i[0] = i[1]-1; if (i[0]<0)        i[0] = 0;
  i[3] = i[2]+1; if (i[3]>dim[0]+1) i[3] = dim[0]+1;
  j[0] = j[1]-1; if (j[0]<0)        j[0] = 0;
  j[3] = j[2]+1; if (j[3]>dim[1]+1) j[3] = dim[1]+1;
  k[0] = k[1]-1; if (k[0]<0)        k[0] = 0;
  k[3] = k[2]+1; if (k[3]>dim[2]+1) k[3] = dim[2]+1;

  for (int ki=0; ki<=3; ki++){
    for (int ji=0; ji<=3; ji++){
      // interp sideways
      fj[ji] = interp(u[ix(i[0],j[ji],k[ki])], u[ix(i[1],j[ji],k[ki])],
		      u[ix(i[2],j[ji],k[ki])], u[ix(i[3],j[ji],k[ki])], s[0]);
    }
    // interp vertically
    fk[ki] = interp(fj[0], fj[1], fj[2], fj[3], s[1]);
  }
  // interp depthwise
  float ret = interp(fk[0], fk[1], fk[2], fk[3], s[2]);  
  
  return(ret);
}

template <class T>
T getVal(float x, float y, float z, const T* u){

  T ret;

  int i[4], j[4], k[4];
  float s[3];
  SbVec3f fj[4], fk[4];
  
  pos2idx(x, y, z, i[1], j[1], k[1], i[2], j[2], k[2], s[0], s[1], s[2]);

  i[0] = i[1]-1; if (i[0]<0)   i[0]=0;
  i[3] = i[2]+1; if (i[3]>dim[0]+1) i[3] = dim[0]+1;
  j[0] = j[1]-1; if (j[0]<0)        j[0] = 0;
  j[3] = j[2]+1; if (j[3]>dim[1]+1) j[3] = dim[1]+1;
  k[0] = k[1]-1; if (k[0]<0)        k[0] = 0;
  k[3] = k[2]+1; if (k[3]>dim[2]+1) k[3] = dim[2]+1;

  for (int ki=0; ki<=3; ki++){
    for (int ji=0; ji<=3; ji++){
      // interp sideways
      fj[ji][0] = interp(u[ix(i[0],j[ji],k[ki])][0], u[ix(i[1],j[ji],k[ki])][0],
			 u[ix(i[2],j[ji],k[ki])][0], u[ix(i[3],j[ji],k[ki])][0], s[0]);
      fj[ji][1] = interp(u[ix(i[0],j[ji],k[ki])][1], u[ix(i[1],j[ji],k[ki])][1],
			 u[ix(i[2],j[ji],k[ki])][1], u[ix(i[3],j[ji],k[ki])][1], s[1]);
      fj[ji][2] = interp(u[ix(i[0],j[ji],k[ki])][2], u[ix(i[1],j[ji],k[ki])][2],
			 u[ix(i[2],j[ji],k[ki])][2], u[ix(i[3],j[ji],k[ki])][2], s[2]);
    }
    // interp vertically
    fk[ki][0] = interp(fj[0][0], fj[1][0], fj[2][0], fj[3][0], s[1]);
    fk[ki][1] = interp(fj[0][1], fj[1][1], fj[2][1], fj[3][1], s[1]);
    fk[ki][2] = interp(fj[0][2], fj[1][2], fj[2][2], fj[3][2], s[1]);
  }
  // interp depthwise
  ret[0] = interp(fk[0][0], fk[1][0], fk[2][0], fk[3][0], s[2]);  
  ret[1] = interp(fk[0][1], fk[1][1], fk[2][1], fk[3][1], s[2]);  
  ret[2] = interp(fk[0][2], fk[1][2], fk[2][2], fk[3][2], s[2]);  

}

// b=0 : boundary same as neighbors
// b=1 : vertical walls negative of neighbors
// b=2 : floor/ceiling negative of neighbors
template <class T>
void setbnd(int b, T *x){

  int nx0 = dim[0], nx1 = dim[0]+1;
  int ny0 = dim[1], ny1 = dim[1]+1;
  int nz0 = dim[2], nz1 = dim[2]+1;

  // walls
  for (int j=1; j<=dim[2]; j++){
    for (int i=1; i<=dim[1]; i++){
      x[ix(0,  i,j)] = b==1 ? -x[ix(1,  i,j)] : x[ix(1,  i,j)];
      x[ix(nx1,i,j)] = b==1 ? -x[ix(nx0,i,j)] : x[ix(nx0,i,j)];
    }
  }
      
  for (int j=1; j<=dim[2]; j++){
    for (int i=1; i<=dim[0]; i++){
      x[ix(i,0,  j)] = b==2 ? -x[ix(i,1,  j)] : x[ix(i,1,  j)];
      x[ix(i,ny1,j)] = b==2 ? -x[ix(i,ny0,j)] : x[ix(i,ny0,j)];
    }
  }
      
  for (int j=1; j<=dim[1]; j++){
    for (int i=1; i<=dim[0]; i++){
      x[ix(i,j,  0)] = b==3 ? -x[ix(i,j,1  )] : x[ix(i,j,1  )];
      x[ix(i,j,nz1)] = b==3 ? -x[ix(i,j,nz0)] : x[ix(i,j,nz0)];
    }
  }

  //   y
  //   |
  //   o--x
  //  /
  // z
  //
  
  // edges
  for (int i=1; i<=dim[0]; i++){
    x[ix(i,  0,  0)] = 0.5f * (x[ix(i,1,  0  )] + x[ix(i,  0,  1  )]);
    x[ix(i,ny1,  0)] = 0.5f * (x[ix(i,ny0,0  )] + x[ix(i,ny1,  1  )]);
    x[ix(i,  0,nz1)] = 0.5f * (x[ix(i,1,nz1  )] + x[ix(i,  0,  nz0)]);
    x[ix(i,ny1,nz1)] = 0.5f * (x[ix(i,ny0,nz1)] + x[ix(i,ny1,  nz0)]);
  }

  for (int i=1; i<=dim[1]; i++){
    x[ix(0,  i,  0)] = 0.5f * (x[ix(1,i,  0  )] + x[ix(0,  i,  1  )]);
    x[ix(nx1,i,  0)] = 0.5f * (x[ix(nx0,i,0  )] + x[ix(nx1,i,  1  )]);
    x[ix(0,  i,nz1)] = 0.5f * (x[ix(1,i,nz1  )] + x[ix(0,  i,  nz0)]);
    x[ix(nx1,i,nz1)] = 0.5f * (x[ix(nx0,i,nz1)] + x[ix(nx1,i,  nz0)]);
  }

  for (int i=1; i<=dim[2]; i++){
    x[ix(0,  0,  i)] = 0.5f * (x[ix(0,1,  i  )] + x[ix(1,  0,  i)]);
    x[ix(0,ny1,  i)] = 0.5f * (x[ix(0,ny0,i  )] + x[ix(1,ny1,  i)]);
    x[ix(nx1,  0,i)] = 0.5f * (x[ix(nx1,1,i  )] + x[ix(nx0,  0,i)]);
    x[ix(nx1,ny1,i)] = 0.5f * (x[ix(nx1,ny0,i)] + x[ix(nx0,ny1,i)]);
  }
  

  // corners
  x[ix(0,0    ,0)] = (float)(1./3.)*(x[ix(0,1,    0)] + x[ix(1,0,  0)] + x[ix(0,0,      1)]);
  x[ix(nx1,0  ,0)] = (float)(1./3.)*(x[ix(nx0,0,  0)] + x[ix(nx1,1,0)] + x[ix(nx1,0,    1)]);
  x[ix(0,ny1  ,0)] = (float)(1./3.)*(x[ix(0,ny0,  0)] + x[ix(1,ny1,0)] + x[ix(0,ny1,    1)]);
  x[ix(nx1,ny1,0)] = (float)(1./3.)*(x[ix(nx1,ny0,0)] + x[ix(nx0,ny1,0)] + x[ix(nx1,ny1,1)]);

  x[ix(0,0    ,nz1)] = (float)(1./3.)*(x[ix(0,1,  nz1  )] + x[ix(1,0,  nz1  )] + x[ix(0,0,    nz0)]);
  x[ix(nx1,0  ,nz1)] = (float)(1./3.)*(x[ix(nx0,0,  nz1)] + x[ix(nx1,1,nz1  )] + x[ix(nx1,0,  nz0)]);
  x[ix(0,ny1  ,nz1)] = (float)(1./3.)*(x[ix(0,ny0,  nz1)] + x[ix(1,ny1,nz1  )] + x[ix(0,ny1,  nz0)]);
  x[ix(nx1,ny1,nz1)] = (float)(1./3.)*(x[ix(nx1,ny0,nz1)] + x[ix(nx0,ny1,nz1)] + x[ix(nx1,ny1,nz0)]);


  // objects
  int i0, j0, k0, i1, j1, k1, h, w, l;
  w = MAX2(1,(int)(.1*length[1]*invvoxel[0])); if (w%2) w++;
  h = MAX2(1,(int)(.1*length[1]*invvoxel[1])); if (h%2) h++;
  l = MAX2(1,(int)(.1*length[1]*invvoxel[2])); if (l%2) l++;

  i0 = dim[0]/2 - w/2     + 1;   i1 = i0+w-1;
  j0 = dim[1]/2 - h/w     + 1;   j1 = j0+h-1;
  k0 = dim[2]/2 - l/2     + 1;   k1 = k0+l-1;

  if (i0<1) i0 = 1; if (i1>dim[0]) i1 = dim[0];
  if (j0<1) j0 = 1; if (j1>dim[1]) j1 = dim[1];
  if (k0<1) k0 = 1; if (k1>dim[2]) k1 = dim[2];
	 
  // initialize explosion
  for (int k=k0; k<=k1; k++)
    for (int j=j0; j<=j1; j++)
      for (int i=i0; i<=i1; i++){
	if (b==0){
	  int dx = (i-i0)-(w/2);
	  int dy = (j-j0)-(h/2);
	  int dz = (k-k0)-(l/2);
	}
      }
  

  
}


// Gauss-Seidel relaxation,
// from "Real-Time Fluid Dynamics for Games" by Jos Stam
void threaded_lin_solve(int b, float *x, float *x0, float a, float c);
void lin_solve(int b, float *x, float *x0, float a, float c){

  // call threaded version
  threaded_lin_solve(b,x,x0,a,c);
  return;

  // 20 iterations
  for (int m=0; m<20; m++){

    // solve x0 = x - a*[sum(neighbor x) - d*x)] for x
    // x + a*d*x = x0 + a*[sum(neighbor x)]
    // x = (x0 + a*[sum(neighbor x)])/(1+a*d)
    // x = (x0 + a*[sum(neighbor x)])/c    where c = 1+a*d
    for (int k=1; k<=dim[2]; k++){
      for (int j=1; j<=dim[1]; j++){
	for (int i=1; i<=dim[0]; i++){
	  x[ix(i,j,k)] = (x0[ix(i,j,k)] + a*(x[ix(i+1,j,k)] + x[ix(i-1,j,k)] +
					     x[ix(i,j+1,k)] + x[ix(i,j-1,k)] +
					     x[ix(i,j,k+1)] + x[ix(i,j,k-1)] )) / c;
	}
      }
    }
    setbnd(b, x);
  }
}



pthread_mutex_t mutex    = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  cond     = PTHREAD_COND_INITIALIZER;

int numDone = 0;

typedef struct lin_solve_t {
  int id;
  int b;
  float *x, *x0, a, c;
} lin_solve_t;

void* lin_solve_chunk(void* ctx_);

void threaded_lin_solve(int b, float *x, float *x0, float a, float c){
  
  pthread_t thread[numThreads];
  lin_solve_t ctx[numThreads];
    
  for (int i=0; i<numThreads; i++){
    ctx[i].id = i;
    ctx[i].b = b;
    ctx[i].x = x;
    ctx[i].x0 = x0;
    ctx[i].a = a;
    ctx[i].c = c;
    pthread_create(&thread[i], NULL, lin_solve_chunk, (void*)&ctx[i]);
  }
  
  for (int i=0; i<numThreads; i++)
    pthread_join(thread[i], NULL);
}

// Gauss-Seidel relaxation,
// from "Real-Time Fluid Dynamics for Games" by Jos Stam
void* lin_solve_chunk(void* ctx_){

  lin_solve_t *ctx = (lin_solve_t *)ctx_;
  int b = ctx->b;
  float *x = ctx->x;
  float *x0 = ctx->x0;
  float a = ctx->a;
  float c = ctx->c;

  //(n/t+1)*id ... (n/t+1)*(id+1)
  const int kpt = dim[2]/numThreads+1;
  
  // 20 iterations
  for (int m=0; m<20; m++){

    // solve x0 = x - a*[sum(neighbor x) - d*x)] for x
    // x + a*d*x = x0 + a*[sum(neighbor x)]
    // x = (x0 + a*[sum(neighbor x)])/(1+a*d)
    // x = (x0 + a*[sum(neighbor x)])/c    where c = 1+a*d
//    for (int k=ctx->id+1; k<=n; k+=numThreads){
    for (int k=MAX2(1,ctx->id*kpt); k<MIN2(dim[2]+1,(ctx->id+1)*kpt); k++){
      for (int j=1; j<=dim[1]; j++){
	for (int i=1; i<=dim[2]; i++){
	  x[ix(i,j,k)] = (x0[ix(i,j,k)] + a*(x[ix(i+1,j,k)] + x[ix(i-1,j,k)] +
					     x[ix(i,j+1,k)] + x[ix(i,j-1,k)] +
					     x[ix(i,j,k+1)] + x[ix(i,j,k-1)] )) / c;
	}
      }
    }


    // sync
    pthread_mutex_lock(&mutex);

    numDone++;

    // wait for threads
    if (numDone == numThreads){
      numDone = 0;
      pthread_cond_broadcast(&cond);
    }
    else
      pthread_cond_wait(&cond, &mutex);

    // do once
    if (ctx->id == 0)
      setbnd(b, x);
 
    pthread_mutex_unlock(&mutex);
    
  }
}

void project(float *u, float *v, float *w, float *u0, float *v0, float *w0){
  float *p = u0;
  float *div = v0;

  // find derivative
  for (int k=1; k<=dim[2]; k++){
    for (int j=1; j<=dim[1]; j++){
      for (int i=1; i<=dim[0]; i++){
	// divergence * facearea = f(+1)-(f-2)/2h * h^2
	div[ix(i,j,k)] = ((u[ix(i+1,j,k)]-u[ix(i-1,j,k)])*voxel[0] +
			  (v[ix(i,j+1,k)]-v[ix(i,j-1,k)])*voxel[1] +
			  (w[ix(i,j,k+1)]-w[ix(i,j,k-1)])*voxel[2]) * -0.5f;
	p[ix(i,j,k)] = 0.0;
      }
    }
  }
  
  if (t<=0.5f){
    
    int i0, j0, k0, i1, j1, k1, h, w, l;
    w = MAX2(1,(int)(.1*length[1]*invvoxel[0])); if (w%2) w++;
    h = MAX2(1,(int)(.1*length[1]*invvoxel[1])); if (h%2) h++;
    l = MAX2(1,(int)(.1*length[1]*invvoxel[2])); if (l%2) l++;

    i0 = dim[0]/2 - w/2     + 1;   i1 = i0+w-1;
    j0 = 1;                        j1 = j0+h-1;
    k0 = dim[2]/2 - l/2     + 1;   k1 = k0+l-1;

    if (i0<1) i0 = 1; if (i1>dim[0]) i1 = dim[0];
    if (j0<1) j0 = 1; if (j1>dim[1]) j1 = dim[1];
    if (k0<1) k0 = 1; if (k1>dim[2]) k1 = dim[2];
    int tot = (i1-i0+1)*(j1-j0+1)*(k1-k0+1);
    //cerr << "tot=" << tot << " i=" << (i1-i0+1) << "\n";
	 
    // initialize explosion
    for (int k=k0; k<=k1; k++)
      for (int j=j0; j<=j1; j++)
	for (int i=i0; i<=i1; i++){
 	  div[ix(i,j,k)] = /*0.005*/tempFloat/(float)((i1-k0+1)*(i1-k0+1));
	}
  }
  
  setbnd(0, div);
  setbnd(0, p);
  
  lin_solve(0, p, div, 1, 6);

  // subtract to create mass conserving field
  for (int k=1; k<=dim[2]; k++){
    for (int j=1; j<=dim[1]; j++){
      for (int i=1; i<=dim[0]; i++){
	u[ix(i,j,k)] -= 0.5f*invvoxel[0]*(p[ix(i+1,j,k)] - p[ix(i-1,j,k)]);
	v[ix(i,j,k)] -= 0.5f*invvoxel[1]*(p[ix(i,j+1,k)] - p[ix(i,j-1,k)]);
	w[ix(i,j,k)] -= 0.5f*invvoxel[2]*(p[ix(i,j,k+1)] - p[ix(i,j,k-1)]);
      }
    }
  }
  
  setbnd(1, u);
  setbnd(2, v);
  setbnd(3, w);
}

void diffuse(int b, float *x, float *x0){
  float a = dt*visc*dim[b-1]*dim[b-1];
  lin_solve(b, x, x0, a, 1+6*a);
}

inline void swap(float *&x, float *&x0){
  float *t = x;
  x = x0;
  x0 = t;
}

								
void stable_solve(float *u, float *v, float *w,
		  float *u0, float *v0, float *w0,
		  float visc, float dt){

  int i, j, k, i0, j0, k0, i1, j1, k1;

  // add forces times timestep
  for (k=1; k<=dim[2]; k++){
    for (j=1; j<=dim[1]; j++){
      for (i=1; i<=dim[0]; i++){
	u[ix(i,j,k)] += dt*u0[ix(i,j,k)];
	v[ix(i,j,k)] += dt*v0[ix(i,j,k)];
	w[ix(i,j,k)] += dt*w0[ix(i,j,k)];
      }
    }
  }
  setbnd(1, u);
  setbnd(2, v);
  setbnd(3, w);

  // add particles and temperature
  if (t<=0.5f){

    // add particles
    addParticles((int)(desiredNumParticles/(0.5f/dt)));    

    int i0, j0, k0, i1, j1, k1, h, w, l;
    w = MAX2(1,(int)(.1*length[1]*invvoxel[0])); if (w%2) w++;
    h = MAX2(1,(int)(.1*length[1]*invvoxel[1])); if (h%2) h++;
    l = MAX2(1,(int)(.1*length[1]*invvoxel[2])); if (l%2) l++;

    i0 = dim[0]/2 - w/2     + 1;   i1 = i0+w-1;
    j0 = 1;                        j1 = j0+h-1;
    k0 = dim[2]/2 - l/2     + 1;   k1 = k0+l-1;

    if (i0<1) i0 = 1; if (i1>dim[0]) i1 = dim[0];
    if (j0<1) j0 = 1; if (j1>dim[1]) j1 = dim[1];
    if (k0<1) k0 = 1; if (k1>dim[2]) k1 = dim[2];

    // initialize explosion
    float f = t/0.5f;
    f = powf(f, heatExponent);

    for (int k=k0; k<=k1; k++)
      for (int j=j0; j<=j1; j++)
	for (int i=i0; i<=i1; i++)
	  temp[ix(i,j,k)] = ambientTemp+(maxTemp-ambientTemp)*f;
    
    setbnd(0, temp);
  }
   

  swap(u, u0);
  swap(v, v0);
  swap(w, w0);
  diffuse(1, u, u0);
  diffuse(2, v, v0);
  diffuse(3, w, w0);
  
  project(u, v, w, u0, v0, w0);

  swap(u, u0);
  swap(v, v0);
  swap(w, w0);
  
  // self-advect using semi-Lagrangian
  float x, y, z, x0, y0, z0;
  for (z=0.5f*voxel[2], k=1; k<=dim[2]; k++,z+=voxel[2]){
    for (y=0.5f*voxel[1], j=1; j<=dim[1]; j++,y+=voxel[1]){
      for (x=0.5f*voxel[0], i=1; i<=dim[0]; i++,x+=voxel[0]){

	x0 = x-dt*u0[ix(i,j,k)];
	y0 = y-dt*v0[ix(i,j,k)];
	z0 = z-dt*w0[ix(i,j,k)];

	u[ix(i,j,k)] = getVal(x0, y0, z0, u0);
	v[ix(i,j,k)] = getVal(x0, y0, z0, v0);
	w[ix(i,j,k)] = getVal(x0, y0, z0, w0);

      }
    }
  }
  setbnd(1, u);
  setbnd(2, v);
  setbnd(3, w);

  project(u, v, w, u0, v0, w0);
  
}


ostream& operator<< (ostream& os, const SbVec3f& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(" << x[0] << ", " << x[1] << ", " << x[2] << ")";
  os << resetiosflags(ios::floatfield) << setprecision(6);
  return(os);
}


void mouseEventHandler(void *data, SoEventCallback *callBack){

  SoEventCallback *eventCB = (SoEventCallback *) data;
  const SoEvent *event = callBack->getEvent();

  static int mouseDown[3];
  static int track = 0;
  static SbVec3f lastPos = SbVec3f(0.,0.,0.);
  static int skip=0;
  
  if (SO_MOUSE_PRESS_EVENT(event, BUTTON1)){
    mouseDown[0] = 1;
    skip = 1;
  }
  else if (SO_MOUSE_RELEASE_EVENT(event, BUTTON1)){
    mouseDown[0] = 0;
  }

  if (SO_MOUSE_PRESS_EVENT(event, BUTTON2)){
    mouseDown[1] = 1;
  }
  else if (SO_MOUSE_RELEASE_EVENT(event, BUTTON2)){
    mouseDown[1] = 0;
  }

  if (mouseDown[0] && mouseDown[1])
    return;

  // add velocity
  if (mouseDown[0]){
    
    const SoPickedPoint *p = eventCB->getPickedPoint();
    if (p){
      
      SbVec3f x = p->getPoint();
      
      if (!skip){
	SbVec3f vec = x-lastPos;
        //cerr << "X :" << x << " Vec: " << vec << "\n";

	int i0, i1, j0, j1, k0, k1;
	float s, t, q;

	pos2idx(x[0], x[1], 0.5f*length[2], i0, j0, k0, i1, j1, k1, s, t, q);
	if (s>0.5f) i0 = i1;
	if (t>0.5f) j0 = j1;
	if (q>0.5f) k0 = k1;

	//cerr << "i=" << i0 << " j=" << j0 << " lastPos: " << lastPos << "\n";
	
	// add force
	float forcemul = 100000.0;
	u0[ix(i0,j0,k0)] += vec[0]*forcemul;
	v0[ix(i0,j0,k0)] += vec[1]*forcemul;
      }
      else
	skip = 0;

      lastPos = x;
    }
  }

  // add density
  if (mouseDown[1]){

    const SoPickedPoint *p = eventCB->getPickedPoint();
    if (p){
      
      SbVec3f x = p->getPoint();

      int i0, i1, j0, j1, k0, k1;
      float s, t, q;

      pos2idx(x[0], x[1], 0.5f*length[2], i0, j0, k0, i1, j1, k1, s, t, q);
      if (s>0.5f) i0 = i1;
      if (t>0.5f) j0 = j1;
      if (q>0.5f) k0 = k1;
    
      dens[ix(i0,j0,k0)] = 1.0;
    }
  }
  
}
				      
void resetTexCoords(){
  for (int j=0; j<dim[1]; j++){
    for (int i=0; i<dim[0]; i++){
      SbVec2f t = SbVec2f((i+0.5f)/dim[0], (j+0.5f)/dim[1]);
      texCoord->point.set1Value(dim[0]*j+i, t);
    }
  }
}


//WW''''''''''''''WWWWW'''''''''''''''''WW''WW''W'''''''''''''WW'WWW'''''''
//WW              WW                    WW  WW  W             WW  WW       
//WW W   WW  W WW WWWW  WW W  WW  WWW  WWWW WW  W  WWW WWW   WWW  WW   WW  
//WWW   WWWW W WW WW    WW W WWWW WW W  WW  WWWWW W WW WW W W WW  WW  WWWW 
//WW W  WW   W WW WW    WW W WW   WW W  WW  WW  W W WW WW W W WW  WW  WW   
//WW  W  WWW  WWW WWWWW  WW   WWW WW W  WW  WW  W  WWW WW W  WWW WWWW  WWW 
//             WW                                                          
//'''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
void keyEventHandler(void *data, SoEventCallback *callBack){
  const SoEvent *event = callBack->getEvent();

  // Save density to file
  if (SO_KEY_PRESS_EVENT(event, D)){
    dump(dens, "density", fileNo++);
  }
  
  // Save scene to file
  else if (SO_KEY_PRESS_EVENT(event, W)){

    SoOutput *out = new SoOutput();
    out->openFile(outFile);
    SoWriteAction *wa = new SoWriteAction(out);
    wa->apply(scene);
    out->closeFile();
    cerr << "Dumped scene graph to " << outFile << ".\n";
    delete out;
    delete wa;
  }

  // pause
  if (SO_KEY_PRESS_EVENT(event, SPACE)){
    animPause = !animPause;
    if (animPause)
      cerr << "Paused simulation.\n";
    else
      cerr << "Continuing simulation.\n";
  }

  // reset simulation
  else if (SO_KEY_PRESS_EVENT(event, R)){
    resetSim();
    cerr << "Reset simulation.\n";
  }
  
  // reset texture coordinates
  else if (SO_KEY_PRESS_EVENT(event, T)){
    resetTexCoords();
    cerr << "Reset texture coordinates.\n";
  }

  // show particles
  else if (SO_KEY_PRESS_EVENT(event, P)){
    showParticles = !showParticles;
    if (showParticles){
      cerr << "Showing particles.\n";
      partSwitch->whichChild.setValue(SO_SWITCH_ALL);
    }
    else {
      cerr << "Hiding particles.\n";
      partSwitch->whichChild.setValue(SO_SWITCH_NONE);
    }
  }
  
  // show material
  else if (SO_KEY_PRESS_EVENT(event, M)){

    showItem = (showItem+1)%showItems;
    
    switch (showItem){

    case SHOW_FORCE:
      cerr << "\nDisplaying force.\n";
      texSwitch->whichChild.setValue(-1);
      break;
      
    case SHOW_SPEED:
      cerr << "\nDisplaying speed.\n";
      texSwitch->whichChild.setValue(-1);
      break;
      
    case SHOW_DENSITY:
      cerr << "\nDisplaying density.\n";
      texSwitch->whichChild.setValue(-1);
      break;
      
    case SHOW_TEXTURE:
      cerr << "\nDisplaying texture.\n";
      resetTexCoords();
      texSwitch->whichChild.setValue(0);
      break;
      
    default:
      break;
    }

  }

}


void animate(void *, SoSensor *){

  // stop simulation
  if (t>=maxTime && renderQuit)
    exit(0);
  else if (animPause || t>=maxTime)
    return;

  // timing variables
  static struct timeval start, end;
  static int initStart = 1;
  if (initStart){
    gettimeofday(&start, NULL);
    initStart = 0;
  }
  
  // advance time
  t += dt;
  
  // advance fluid
  TIME(stable_solve(u, v, w, u0, v0, w0, visc, dt));
    
  // add smoke, temperature
  if (t<=0.5f/*dt*/) addSmoke();
  const float puffTime = 0.5;
  const int numPuffs = 10;//30;
  const float puffsPerSec = numPuffs/puffTime;
  const float timePerPuff = 1.0/puffsPerSec;
  static int lastPuff = 0;
  static int puffs = 0;
  int desiredPuffs = (int)(MIN2(t, puffTime) / timePerPuff + 0.5f);
  //int desiredPuffs = (int)(MIN2(t, t) / timePerPuff + 0.5f);
  while (0 && puffs < desiredPuffs){
    SbVec3f offset(drand48()*2.0-1.0, drand48(), drand48()*2.0-1.0);
    addPuff(SbVec3f(.5,.0,.5)+offset*.1, .07*drand48()+0.05);
    puffs++;
    cerr << "Puff!\n";
//     addPuff(SbVec3f(.5,.05,.5), .1);
//     addPuff(SbVec3f(.55,.03,.5), .05);
  }


  // write density
  if (writeDump){
    static int fileNum=0;
    static int lastDiv = -1;
    const int fps = 30;//10;
    float tpf = 1.0/fps;
    int div = (int)(t/tpf);
    
    if (div != lastDiv){
      char name[100];
      sprintf(name, "%s.dens", dumpName);
      dump(dens, name, fileNum);
      sprintf(name, "%s.temp", dumpName);
      dump(temp, name, fileNum);
      fileNum++;
      lastDiv = div;
    }
  }
  
  
  // reset forces
  for (int k=1; k<=dim[2]; k++){
    for (int j=1; j<=dim[1]; j++){
      for (int i=1; i<=dim[0]; i++){

	const float grav      = -9.8;                        // m/s^2    (earth's gravity)
	const float bouy      = bouyMult;                    // m/(Ks^2) 
	const float voxelVol  = voxel[0]*voxel[1]*voxel[2];  // m^3
	const float airDens   = 1.29;                        // kg/m^3   (from web)
	const float smokeDens = dens[ix(i,j,k)];             // kg/m^3

	// f = ma = mg
	// f_air   =  dens_air            *vol*g
	// f_smoke = (dens_air+dens_smoke)*vol*g
	// f = f_smoke - f_air = dens_smoke*vol*g
	// a = f/m = dens_smoke*vol*g / ((dens_air+dens_smoke)*vol)
	//         = dens_smoke/(dens_air+dens_smoke) * g
	
	SbVec3f f(0,0,0);
	f[1] =                                     // acceleration, m/s^2
	  smokeDens/(airDens+smokeDens)*grav +     // gravity
	  bouy*(temp[ix(i,j,k)]-ambientTemp);      // bouyancy

	//f[1] = (smokeDens>0.0) ? 10.0 : 0.0;

	u0[ix(i,j,k)] = f[0];
	v0[ix(i,j,k)] = f[1];
	w0[ix(i,j,k)] = f[2];
      }
    }
  }

  
  // increase vorticity
  // using either vorticity confinement or vortex particles method

  // vorticity confinement
  if (!useVorParticles) {  

    // compute vorticity
    for (int k=1; k<=dim[2]; k++){
      for (int j=1; j<=dim[1]; j++){
	for (int i=1; i<=dim[0]; i++){

	  //  vort = grad x U
	  //
	  //  |  i     j    k  |
	  //  | d/dx d/dy d/dz |
	  //  |  u     v    w  |
      
	  vortg[ix(i,j,k)][0] = ((w[ix(i,j+1,k)] - w[ix(i,j-1,k)])*invvoxel[2] -
				 (v[ix(i,j,k+1)] - v[ix(i,j,k-1)])*invvoxel[1]) * 0.5f;
	  vortg[ix(i,j,k)][1] = ((u[ix(i,j,k+1)] - u[ix(i,j,k-1)])*invvoxel[0] -
				 (w[ix(i+1,j,k)] - w[ix(i-1,j,k)])*invvoxel[2]) * 0.5f;
	  vortg[ix(i,j,k)][2] = ((v[ix(i+1,j,k)] - v[ix(i-1,j,k)])*invvoxel[1] -
				 (u[ix(i,j+1,k)] - u[ix(i,j-1,k)])*invvoxel[0]) * 0.5f;
	}
      }
    }
    setbnd(0, vortg);

    // add vorticity confinement
    for (int k=1; k<=dim[2]; k++){
      for (int j=1; j<=dim[1]; j++){
	for (int i=1; i<=dim[0]; i++){

	  SbVec3f f, eta;
	  // n = grad |w|
	  eta[0] =
	    (vortg[ix(i+1,j,k)].length() - vortg[ix(i-1,j,k)].length()) *
	    0.5f*invvoxel[0];
	  eta[1] =
	    (vortg[ix(i,j+1,k)].length() - vortg[ix(i,j-1,k)].length()) *
	    0.5f*invvoxel[1];
	  eta[2] =
	    (vortg[ix(i,j,k+1)].length() - vortg[ix(i,j,k-1)].length()) *
	    0.5f*invvoxel[2];
	  eta.normalize();
     
	  f = vortEps*(eta.cross(vortg[ix(i,j,k)]));
	  f[0] *= voxel[0];
	  f[1] *= voxel[1];
	  f[2] *= voxel[2];
      
	  u0[ix(i,j,k)] += f[0];
	  v0[ix(i,j,k)] += f[1];
	  w0[ix(i,j,k)] += f[2];
	}
      }
    }
  }
  
  else { // vortex particle method
  	
    // compute velocity derivatives
    for (int k=1; k<=dim[2]; k++){
      for (int j=1; j<=dim[1]; j++){
	for (int i=1; i<=dim[0]; i++){
	  
	  du[ix(i,j,k)][0] = (u[ix(i+1,j,k)]-u[ix(i-1,j,k)])*0.5f*invvoxel[0];
	  du[ix(i,j,k)][1] = (u[ix(i,j+1,k)]-u[ix(i,j-1,k)])*0.5f*invvoxel[1];
	  du[ix(i,j,k)][2] = (u[ix(i,j,k+1)]-u[ix(i,j,k-1)])*0.5f*invvoxel[2];

	  dv[ix(i,j,k)][0] = (v[ix(i+1,j,k)]-v[ix(i-1,j,k)])*0.5f*invvoxel[0];
	  dv[ix(i,j,k)][1] = (v[ix(i,j+1,k)]-v[ix(i,j-1,k)])*0.5f*invvoxel[1];
	  dv[ix(i,j,k)][2] = (v[ix(i,j,k+1)]-v[ix(i,j,k-1)])*0.5f*invvoxel[2];

	  dw[ix(i,j,k)][0] = (w[ix(i+1,j,k)]-w[ix(i-1,j,k)])*0.5f*invvoxel[0];
	  dw[ix(i,j,k)][1] = (w[ix(i,j+1,k)]-w[ix(i,j-1,k)])*0.5f*invvoxel[1];
	  dw[ix(i,j,k)][2] = (w[ix(i,j,k+1)]-w[ix(i,j,k-1)])*0.5f*invvoxel[2];
	}
      }
    }
    setbnd(0, du);
    setbnd(0, dv);
    setbnd(0, dw);

    // update particle vorticities
    for (int i=0; i<numParticles; i++){
    
      // particle position
      SoTransform *trans = (SoTransform *) partSeps[i]->getChild(0);
      SbVec3f pos = trans->translation.getValue();
    
      // interpolate derivatives
      // fedkiw uses trilinear
      SbVec3f dup = getValLinear(pos[0], pos[1], pos[2], du);
      SbVec3f dvp = getValLinear(pos[0], pos[1], pos[2], dv);
      SbVec3f dwp = getValLinear(pos[0], pos[1], pos[2], dw);

      // vort += (w . grad) u
      SbVec3f vort(0.0, 0.0, 0.0);
      vort[0] = vortp[i][0]*dup[0] + vortp[i][1]*dup[1] + vortp[i][2]*dup[2];
      vort[1] = vortp[i][0]*dvp[0] + vortp[i][1]*dvp[1] + vortp[i][2]*dvp[2];
      vort[2] = vortp[i][0]*dwp[0] + vortp[i][1]*dwp[1] + vortp[i][2]*dwp[2];

      vortp[i] += dt*vort;

      // normalize vorticity
      vortp[i].normalize();
      vortp[i] *= 2.e-3f;
    
    }

    // add particles vorticity
    for (int m=0; m<numParticles; m++){

      // particle position
      SoTransform *trans = (SoTransform *) partSeps[m]->getChild(0);
      SbVec3f pos = trans->translation.getValue();
    
      float rad = dim[0]/45.f+0.1;//n/64.f;
      int radi = (int)rad;
      if (rad-radi>0.f || radi<=1) radi++;
      rad *= voxel[0];

      int i0, j0, k0;
      pos2idx(pos, i0, j0, k0);

      for (int k=MAX2(1,k0-radi); k<MIN2(dim[2],k0+radi); k++){
	for (int j=MAX2(1,j0-radi); j<MIN2(dim[1],j0+radi); j++){
	  for (int i=MAX2(1,i0-radi); i<MIN2(dim[0],i0+radi); i++){

	    SbVec3f x = idx2pos(i, j, k);
	    SbVec3f wp = gaussFilt(x, pos, rad) * vortp[m];
	    SbVec3f Np = pos-x;
	    Np.normalize();
	    SbVec3f Fp = vortEps * Np.cross(wp);

	    u0[ix(i,j,k)] += Fp[0];
	    v0[ix(i,j,k)] += Fp[1];
	    w0[ix(i,j,k)] += Fp[2];
	  }
	}
      }
    }
  }


  // advect density and temperature
  swap(dens, dens0);
  swap(temp, temp0);
  
  float x, y, z, x0, y0, z0;
  for (int k=1; k<=dim[2]; k++){
    for (int j=1; j<=dim[1]; j++){
      for (int i=1; i<=dim[0]; i++){

	// voxel center position
	x = origin[0]+voxel[0]*(i-0.5f);
	y = origin[1]+voxel[1]*(j-0.5f);
	z = origin[2]+voxel[2]*(k-0.5f);

	// advected position
	x0 = x-dt*u[ix(i,j,k)];
	y0 = y-dt*v[ix(i,j,k)];
	z0 = z-dt*w[ix(i,j,k)];
	
	dens[ix(i,j,k)] = getValLinear(x0, y0, z0, dens0);

	// temperature advection from Nguyen at al.
	const float cT = 3000.0f; // K/s 
	float newTemp = getValLinear(x0, y0, z0, temp0);
	float ratio = (newTemp-ambientTemp)/(maxTemp-ambientTemp);
	newTemp -= cT*ratio*ratio*ratio*ratio*dt;
	temp[ix(i,j,k)] = newTemp;
      }
    }
  }
  setbnd(0, dens);
  setbnd(0, temp);

  
  if (showItem != SHOW_TEXTURE){

    static float maxForce = 0.0;

    // adjust vertex colors
    for (int k=1; k<=dim[2]; k++){

      SbVec2s size;
      int nc;
      unsigned char *tex = planeTexs[k-1]->image.startEditing(size, nc);
      
      for (int j=1; j<=dim[1]; j++){
	for (int i=1; i<=dim[0]; i++){

	  SbColor color;
	
	  if (showItem == SHOW_SPEED){
	
	    // speed
	    float speed =
	      u[ix(i,j,k)]*u[ix(i,j,k)] +
	      v[ix(i,j,k)]*v[ix(i,j,k)] +
	      w[ix(i,j,k)]*w[ix(i,j,k)];
	    speed = sqrt(speed);

	    // record speed
	    if (speed>maxSpeed)
	      maxSpeed = speed;

	    float hue = sqrt(speed/maxSpeed);
	    hue = -2.0/3.0*hue+2.0/3.0;

	    color.setHSVValue(hue, 0.8, 0.8);

	  }
	  else if (showItem == SHOW_FORCE){
	  
	    // force
	    float force =
	      u0[ix(i,j,k)]*u0[ix(i,j,k)] +
	      v0[ix(i,j,k)]*v0[ix(i,j,k)] +
	      w0[ix(i,j,k)]*w0[ix(i,j,k)];
	    force = sqrt(force);
	  
	    // record force
	    if (force>maxForce)
	      maxForce = force;
	  
	    float hue = sqrt(force/maxForce);
	    hue = -2.0/3.0*hue+2.0/3.0;
	
	    color.setHSVValue(hue, 0.8, 0.8);
	  }
	  else if (showItem == SHOW_DENSITY){
	  
	    float c = dens[ix(i,j,k)]/maxDensity;
	    //c = temp[ix(i,j,k)]/maxTemp;

	    if (c>1.0)
	      c = 1.0;
	    
	    color.setValue(c,c,c);
	  }
	
	  //planeMat[k-1]->diffuseColor.set1Value((i-1)+(j-1)*n, color);
	  int c = clamp((int)(color[0]*255.0f), 0, 255);
	  tex[((i-1)+(j-1)*dim[0])*3+0] = c;
	  tex[((i-1)+(j-1)*dim[0])*3+1] = c;
	  tex[((i-1)+(j-1)*dim[0])*3+2] = c;
	}
      }
      
      planeTexs[k-1]->image.finishEditing();
      
    }



  }

  else {

    // advect texture coordinates
    int k = dim[2]/2;
    for (int j=1; j<=dim[1]; j++){
      for (int i=1; i<=dim[0]; i++){

	// voxel center position
	x = origin[0]+voxel[0]*(i-0.5f);
	y = origin[1]+voxel[1]*(j-0.5f);
	z = origin[2]+voxel[2]*(k-0.5f);

	// advected position
	x0 = x-dt*u[ix(i,j,k)];
	y0 = y-dt*v[ix(i,j,k)];
	z0 = z;

	// FIXME
//	SbVec2f newCoord = getVal(x0, y0, texCoord->point.getValues(0));
	
//	texCoord0->point.set1Value(i+n*j, newCoord);
      }
    }

    // copy texture coordinates
    SbVec2f *coord = texCoord->point.startEditing();
    SbVec2f *coord0 = texCoord0->point.startEditing();
    for (int i=0; i<dim[0]*dim[1]; i++)
      coord[i] = coord0[i];
    texCoord->point.finishEditing();
    texCoord->point.finishEditing();
    texCoord0->point.finishEditing();

  }

  
  // advect particles
  if (1||showParticles) {
    for (int i=0; i<numParticles; i++){

      SoTransform *trans = (SoTransform *) partSeps[i]->getChild(0);
      SoMaterial *mat = (SoMaterial *) partSeps[i]->getChild(1);

      // particle orientation
      SbVec3f pv(vortp[i]);
      pv.normalize();
      trans->rotation.setValue(SbRotation(SbVec3f(0,1,0), pv));
    
      // particle position
      SbVec3f pos = trans->translation.getValue();

      SbVec3f u_s;
      u_s[0] = getValLinear(pos[0], pos[1], pos[2], u);
      u_s[1] = getValLinear(pos[0], pos[1], pos[2], v);
      u_s[2] = getValLinear(pos[0], pos[1], pos[2], w);

      // advected position
      pos[0] += dt*u_s[0];
      pos[1] += dt*u_s[1];
      pos[2] += dt*u_s[2];

      // wrap particles
      for (int j=0; j<3; j++){
	if (pos[j] > origin[j]+length[j])
	  pos[j] = origin[j]+length[j];
	else if (pos[j] < origin[j])
	  pos[j] = origin[j];
      }

      trans->translation.setValue(pos);

      // record speed
      float s = u_s.length();
      if (s>maxSpeed)
	maxSpeed = s;

      // set color
      float hue = sqrt(s/maxSpeed);    // 0->1
      hue = -2.0/3.0*hue+2.0/3.0;      // 0.3333->1.0
      mat->diffuseColor.setHSVValue(hue, 1.0, 0.8);

      // show particle vorticicty
      //SbVec3f v = vortp[i];
      //v.normalize();
      //showRay(pos, v, .1);
    }
    //animPause = 1;
  }

  // progress readout
  gettimeofday(&end, NULL);
  int ms = delay(start, end);
  cerr << "\r" << t << " ("
       << ms << " ms / "
       << ms/1000. << " s) "
       << "       ";
  gettimeofday(&start, NULL);

}

void addSmoke(){

  int i0, j0, k0, i1, j1, k1, h, w, l;
  w = MAX2(1,(int)(.1*length[1]*invvoxel[0])); if (w%2) w++;
  h = MAX2(1,(int)(.1*length[1]*invvoxel[1])); if (h%2) h++;
  l = MAX2(1,(int)(.1*length[1]*invvoxel[2])); if (l%2) l++;

  i0 = dim[0]/2 - w/2     + 1;   i1 = i0+w-1;
  j0 = 1;                        j1 = j0+h-1;
  k0 = dim[2]/2 - l/2     + 1;   k1 = k0+l-1;

  // http://www.dieselnet.com/calculator/smoke1.html	  
  const float newDens = 8.0e-3; // kg/m^3
  

  if (newDens>maxDensity)
    maxDensity = newDens;

  // initialize smoke
  for (int k=MAX2(1,k0); k<=MIN2(dim[2],k1); k++)
    for (int j=MAX2(1,j0); j<=MIN2(dim[1],j1); j++)
      for (int i=MAX2(1,i0); i<=MIN2(dim[0],i1); i++){
	dens[ix(i,j,k)] = newDens;
        //0.8; //*= sqrt(drand48())*2.0;//1.0;
	//temp[ix(i,j,k)] = 200;
      }

//   s = MAX2(1,w/2);//(int)(.08*n));
//   if (s%2) s++;
//   h = w = l = s;
//   i0 = n/2 - w/2 +1;        i1 = i0+w;
//   j0 = (int)(-.25*h) + 1;   j1 = j0+h;
//   k0 = n/2 - l/2 +1;        k1 = k0+l;

  float maxD = SbVec3f(w/2, j1, l/2).length();
  
  // initialize temp
//   for (int k=MAX2(1,k0); k<=MIN2(n,k1); k++)
//     for (int j=MAX2(1,j0); j<=MIN2(n,j1); j++)
//       for (int i=MAX2(1,i0); i<=MIN2(n,i1); i++){
// 	float d = SbVec3f(i-n/2, j, k-n/2).length();
//  	if (d<maxD){
// 	  float f = 1-d/maxD;
// 	  f *= f;
// 	  temp[ix(i,j,k)] = 100 * f + ambientTemp;
// 	}
//       }

  setbnd(0, dens);
  setbnd(0, temp);
}

void addPuff(SbVec3f x, float r){

  int i0, i1, j0, j1, k0, k1;
  float s, t, q;
  int ri = MAX2(1, (int)(r*dim[0]));
  pos2idx(x[0], x[1], x[2], i0, j0, k0, i1, j1, k1, s, t, q);
  
  // initialize smoke
  for (int k=MAX2(1,k0-ri); k<=MIN2(dim[2],k0+ri); k++)
    for (int j=MAX2(1,j0-ri); j<=MIN2(dim[1],j0+ri); j++)
      for (int i=MAX2(1,i0-ri); i<=MIN2(dim[0],i0+ri); i++){
	SbVec3f x2 = idx2pos(i,j,k);
	float d = (x-x2).length();
	if (d<r){
	  dens[ix(i,j,k)] += 0.8+drand48()*0.4*100.0;
	  temp[ix(i,j,k)] += 200.0+100.0*drand48()*(1.0f-d/r);
	}
      }

  setbnd(0, dens);
  setbnd(0, temp);
}


void resetSim(){

  t = 0;

  // initialize velocities, force, density
  for (int k=1; k<=dim[2]; k++){
    for (int j=1; j<=dim[1]; j++){
      for (int i=1; i<=dim[0]; i++){
	u[ix(i,j,k)] = 0.0;
	v[ix(i,j,k)] = 0.0;
	w[ix(i,j,k)] = 0.0;
	u0[ix(i,j,k)] = 0.0;
	v0[ix(i,j,k)] = 0.0;
	w0[ix(i,j,k)] = 0.0;
	dens[ix(i,j,k)] = dens0[ix(i,j,k)] = 0.0;
	temp[ix(i,j,k)] = temp0[ix(i,j,k)] = ambientTemp;
	vortg[ix(i,j,k)] = SbVec3f(0.,0.,0.);
      }
    }
  }
  setbnd(1, u);
  setbnd(2, v);
  setbnd(3, w);
  setbnd(0, dens);
  setbnd(0, temp);
  setbnd(0, vortg);


//  addParticles(desiredNumParticles);
}


void addParticles(int addNum){

  int startNum = numParticles;
  numParticles += addNum;


  int i0, j0, k0, i1, j1, k1, h, w, l;
  w = MAX2(1,(int)(.1*length[1]*invvoxel[0])); if (w%2) w++;
  h = MAX2(1,(int)(.1*length[1]*invvoxel[1])); if (h%2) h++;
  l = MAX2(1,(int)(.1*length[1]*invvoxel[2])); if (l%2) l++;

  i0 = dim[0]/2 - w/2     + 1;   i1 = i0+w-1;
  j0 = 1;                        j1 = j0+h-1;
  k0 = dim[2]/2 - l/2     + 1;   k1 = k0+l-1;
  SbVec3f min, max;
  min.setValue(voxel[0]*(i0-1),
	       voxel[1]*(j0-1),
	       voxel[2]*(k0-1));
  max.setValue(voxel[0]*(i1),
	       voxel[1]*(j1),
	       voxel[2]*(k1));
	       
  //idx2pos(i0, j0, k0, min);
  //idx2pos(i1, j1, k1, max);

  /*
  // reset particle positions
  int b[3];
  if (numParticles > 0){
    b[0] = (int)(pow((float)numParticles, 1.f/3.f) + 0.5f);
    b[1] = b[0];
    b[2] = b[1];//numParticles/(b[0]*b[1]);
    cerr << "b: " << b[0] << " " << b[1] << " " << b[2];
  }
  cerr << "   total=" << b[0]*b[1]*b[2] << "\n";
  */


  for (int i=startNum; i<numParticles; i++){
    SoSeparator *pSep = new SoSeparator();
    partSeps.push_back(pSep);
    
    SoTransform *trans = new SoTransform();
    pSep->addChild(trans);
				  
    SoMaterial *mat = new SoMaterial();
    mat->diffuseColor.setValue(0.0, 1.0, 0.0);
    pSep->addChild(mat);
    
    //SoSphere *sph = new SoSphere();
    //sph->radius.setValue(0.01);
    //pSep->addChild(sph);
    
    SoCone *cone = new SoCone();
    float size = .01;
    cone->bottomRadius.setValue(size*0.75f);
    cone->height.setValue(size*2.0f);
    pSep->addChild(cone);

    
    partSwitch->addChild(pSep);
//  }

//   for (int m=0; m<numParticles; m++){

//     // particle position
//     SoTransform *trans = (SoTransform *) partSeps[m]->getChild(0);

//     // jittered initial positions
//     int i = m%b[0];
//     int j = ((m-i)/b[0])%b[1];
//     int k = ((m-i-j*b[0])/(b[1]*b[2]))%b[2];
//     const float ofs = 0.5;
//     trans->translation.setValue((i+ofs+1.0*drand48())/b[0]*length[0],
// 				(j+ofs+1.0*drand48())/b[1]*length[1],
// 				(k+ofs+1.0*drand48())/b[2]*length[2] );
//   }
  
  // initialize vorticities
  const float mag = 2.e-3;
  SbVec3f up(0,1,0);
//  for (int i=0; i<numParticles; i++){
//    SoTransform *trans = (SoTransform *) partSeps[i]->getChild(0);
  trans->translation.setValue(min[0]+(max[0]-min[0])*drand48(),
			      min[1]+(max[1]-min[1])*drand48(),
			      min[2]+(max[2]-min[2])*drand48());

  const SbVec3f center = length*0.5f;
    SbVec3f pos = trans->translation.getValue() - center;
    pos[1] = 0.0;
    pos.normalize();

    float theta = 0.0;
    if (tempInt == 0)
       theta = 0.0;
    else if (tempInt == 1)
       theta = M_PI;
    else if (tempInt == 2){
       if (drand48()<0.5f) theta = M_PI;
    }
    else if (tempInt == 3)
       theta = 2.0f*M_PI*drand48();

    SbVec3f v(cos(theta), sin(theta), 0.0f);
    SbRotation rot(SbVec3f(0,0,-1), pos);
    rot.multVec(v, v);
    
    //vortp[i] = v*mag;
    vortp.push_back(v*mag);
  }
}

void dump(float *x, char *name, int t){

  char name2[100];
  sprintf(name2, "dumpfiles/%s.%03d.dump", name, t);
  FILE *out = fopen(name2, "wb");
  if (out == NULL){
    cerr << "Could not open file " << name2
	 << " for writing: " << strerror(errno) << "\n";
    return;
  }

  // write header
  const int hSize = 512;
  char header[hSize];
  for (int i=0; i<hSize; i++)
    header[i] = '\0';
  sprintf(header, "%d %d %d\n%g %g %g\n%g %g %g\n%d byte header, %d floats\n%s\n",
	  dim[0], dim[1], dim[2], origin[0], origin[1], origin[2],
	  length[0], length[1], length[2], hSize, dim[0]*dim[1]*dim[2], cmdLine.c_str());
  fwrite(header, sizeof(char), hSize, out);

  // write data
  // initialize velocities, force, density
  for (int k=1; k<=dim[2]; k++){
    for (int j=1; j<=dim[1]; j++){
      for (int i=1; i<=dim[0]; i++){
 	float xp = x[ix(i,j,k)];
//	c = clamp((int)c, 0, 255);
//	fwrite(&c, sizeof(char), 1, out);
	fwrite(&xp, sizeof(float), 1, out);
      }
    }
  }

  fclose(out);
  cerr << "Wrote " << name2 << "\n";

}

//'''''''''''''WW'''''
//                    
//WWWW W   WWW WW WWW 
//WW WW W W WW WW WW W
//WW WW W W WW WW WW W
//WW WW W  WWW WW WW W
//                    
//''''''''''''''''''''
//
int main(int argc, char *argv[]){

  // seed RNG
  srand48(3);
  
  // form command line
  for (int i=0; i<argc; i++)
    cmdLine = cmdLine + argv[i] + " ";
  
  // read options
  while (1){
    char c = getopt(argc, argv, "b:d:e:f:hi:j:k:mn:t:T:v:p:f:w:rx:");

    if (c == -1)
      break;

    switch (c){

    case 'h':
       cerr
	 << "Usage: " << argv[0]
	 << " [-d dt] [-h] [-o out.iv] [-n res] [-t texture.jpg] [-v visc]\n"
	 << "\n"
	 << "Advects a texture using a 2D fluid solver.\n"
	 << ""
	 << "\n"
	 << " -b num          bouyance multiple (1.0)\n"
	 << " -d num          time step dt (0.01)\n"
	 << " -e num          epsilon to use in vorticity calcs (1000.0)\n"
	 << " -h              this help message\n"
	 << " -j num          number of threads to use (1)\n"
	 << " -o file         name of output scene file (out.iv)\n"
	 << " -m              use vortex particle method\n"
	 << " -n num          resolution of grid, squared (64)\n"
         << " -p num          number of particles to advect (100)\n"
	 << " -r              no display\n"
	 << " -t file         use texture contained in file (texture.jpg)\n"
	 << " -T num          maximum time to perform simulation with (inf)\n"
	 << " -w file         write dumpfiles to file.000.dump (none)\n"
	 << " -x num          heating exponent (0.0 = constant)\n"
         << " -v num          viscosity of fluid (0.001).\n";
	 
      exit(1);
      break;

    case 'b':
      bouyMult = atof(optarg);
      break;
       
    case 'd':
      dt = atof(optarg);
      break;

    case 'e':
      vortEps = atof(optarg);
      break;
      
    case 'f':
      tempFloat = atof(optarg);
      break;

    case 'i':
      tempInt = atoi(optarg);
      break;
       
    case 'j':
      numThreads = atoi(optarg);
      break;
       
    case 'k':
      tempInt2 = atoi(optarg);
      break;

    case 'm':
      useVorParticles = 1;
      break;
      
    case 'n':
      char *next;
      dim[0] = atoi(optarg);
      next = strchr(optarg, 'x');
      if (next == NULL)
	dim[1] = dim[2] = dim[0];
      else {
	next++;
	dim[1] = atoi(next);
	next = strchr(next, 'x')+1;
	dim[2] = atoi(next);
      }
	
      if (dim[0] < 1 || dim[1] < 1 || dim[2] < 1){
	cerr << "Invalid image dimensions: " << optarg << endl;
	exit(1);
      }

      int max;
      max = MAX2(MAX2(dim[0], dim[1]), dim[2]);
      length = SbVec3f(dim[0], dim[1], dim[2])/max;

      voxel    = SbVec3f(length[0]/dim[0], length[1]/dim[1], length[2]/dim[2]);
      invvoxel = SbVec3f(dim[0]/length[0], dim[1]/length[1], dim[2]/length[2]);
      break;

    case 'p':
      desiredNumParticles = atoi(optarg);
      break;

    case 'r':
      renderQuit = 1;
      break;

    case 't':
      textureFilename = optarg;
      break;

    case 'T':
      maxTime = atof(optarg);
      break;
       
    case 'v':
      visc = atof(optarg);
      cerr << "Set viscosity to " << visc << endl;
      break;

    case 'w':
      dumpName = optarg;
      break;

    case 'x':
      heatExponent = atof(optarg);
      break;

     default:
       break;
    }
  }
  
  while (optind<argc){
    cerr << "Unknown option: " << argv[optind++] << "\n";
  }

  // misc variable setup
  if (strcmp(dumpName, "none") == 0)
    writeDump = 0;
  
  // initialize inventor
  if (renderQuit)
    SoDB::init();
  else
    SoXt::init(argv[0]);

  // create scene
  scene = new SoSeparator();

  // create plane sep
  SoSeparator *planeSep = new SoSeparator();
  scene->addChild(planeSep);

  // add bounding box
  SoSeparator *boundBox = new SoSeparator();
  SoDrawStyle *boxStyle = new SoDrawStyle();
  boxStyle->style.setValue(SoDrawStyle::LINES);
  boundBox->addChild(boxStyle);
  SoTransform *boxTrans = new SoTransform();
  boxTrans->scaleFactor.setValue(length);
  boxTrans->translation.setValue(length/2.0f);
  boundBox->addChild(boxTrans);
  SoMaterial *boxMat = new SoMaterial();
  boxMat->emissiveColor.setValue(1,1,1);
  boundBox->addChild(boxMat);
  SoCube *boxCube = new SoCube();
  boxCube->width.setValue(1);
  boxCube->depth.setValue(1);
  boxCube->height.setValue(1);
  boundBox->addChild(boxCube);
  scene->addChild(boundBox);

  // camera
  //SoOrthographicCamera *camera = new SoOrthographicCamera();
  //camera->position.setValue(origin+length/2.0);
  //scene->addChild(camera);
  
  // complexity
  SoComplexity *comp = new SoComplexity();
  comp->textureQuality.setValue(1.0);
  planeSep->addChild(comp);

  // texture
//   SoTexture2 *texture = new SoTexture2();
//   texture->filename.setValue(textureFilename);
//   texture->model.setValue(SoTexture2::DECAL);

//   // put texture in switch
//   texSwitch = new SoSwitch();
//   texSwitch->addChild(texture);
//   texSwitch->whichChild.setValue(showItem==SHOW_TEXTURE?0:-1);
//   planeSep->addChild(texSwitch);

  // material binding
  SoMaterialBinding *matBind = new SoMaterialBinding();
  matBind->value.setValue(SoMaterialBinding::PER_VERTEX);

  // shape hint
  SoShapeHints *quadHint = new SoShapeHints();
  quadHint->vertexOrdering.setValue(SoShapeHints::COUNTERCLOCKWISE);
  planeSep->addChild(quadHint);
  planeSep->addChild(matBind);
  
  // pick style
  SoPickStyle *pickStyle = new SoPickStyle();
  pickStyle->style.setValue(SoPickStyle::BOUNDING_BOX);
  planeSep->addChild(pickStyle);

  // allocate plane materials
  planeMat = new SoMaterial*[dim[2]];
  planeTexs = new SoTexture2*[dim[2]];
  const int nc = 3;
  SbVec2s size(dim[0],dim[1]);
  unsigned char *black = new unsigned char[dim[0]*dim[1]*nc];
  for (int k=0; k<dim[1]*dim[0]*nc; k++)
    black[k] = 0;
  
  // add k grids
  for (int k=0; k<dim[2]; k++){

    // separator
    SoSeparator *gridSep = new SoSeparator();

    // coordinate
    SoCoordinate3 *coord = new SoCoordinate3();
    coord->point.setNum(4);

    // add grid
    texCoord = new SoTextureCoordinate2();
    texCoord0 = new SoTextureCoordinate2();
    texCoord->point.setNum(4);
    texCoord0->point.setNum(4);
    
    // material
    planeMat[k] = new SoMaterial();
    planeMat[k]->diffuseColor.setNum(4);
    for (int i=0; i<4; i++)
      planeMat[k]->diffuseColor.set1Value(i, SbColor(1.,.8,.8));
    planeMat[k]->ambientColor.setValue(0.,0.,0.);
    planeMat[k]->transparency.setValue(1-10.f/dim[2]);
    gridSep->addChild(planeMat[k]);

    coord->point.set1Value(0, SbVec3f(0.5f*voxel[0],
				      0.5f*voxel[1],
				      voxel[2]*(k+0.5f)));
    coord->point.set1Value(1, SbVec3f(length[0]-0.5f*voxel[0],
				      0.5f*voxel[1],
				      voxel[2]*(k+0.5f)));
    coord->point.set1Value(2, SbVec3f(0.5f*voxel[0],
				      length[1]-0.5f*voxel[1],
				      voxel[2]*(k+0.5f)));
    coord->point.set1Value(3, SbVec3f(length[0]-0.5f*voxel[0],
				      length[1]-0.5f*voxel[1],
				      voxel[2]*(k+0.5f)));
    texCoord->point.set1Value(0, SbVec2f(0.0, 0.0));
    texCoord->point.set1Value(1, SbVec2f(1.0, 0.0));
    texCoord->point.set1Value(2, SbVec2f(0.0, 1.0));
    texCoord->point.set1Value(3, SbVec2f(1.0, 1.0));
    texCoord0->point.set1Value(0, SbVec2f(0.0, 0.0));
    texCoord0->point.set1Value(1, SbVec2f(1.0, 0.0));
    texCoord0->point.set1Value(2, SbVec2f(0.0, 1.0));
    texCoord0->point.set1Value(3, SbVec2f(1.0, 1.0));

    gridSep->addChild(coord);
    gridSep->addChild(texCoord);
    
    // vertices are at grid voxel centers
    // so are texture coordinates
    // grid is O,O->1,1  voxel size = 1/n
//     for (int j=0; j<n; j++){
//       for (int i=0; i<n; i++){
// 	SbVec3f p = origin+SbVec3f(voxel[0]*(i+0.5f), voxel[1]*(j+0.5f), voxel[2]*(k+0.5f));
// //	SbVec2f t = SbVec2f((i+0.5f)/n, (j+0.5f)/n);
// 	coord->point.set1Value(n*j+i, p);
// //	texCoord->point.set1Value(n*j+i, t);
// //	texCoord0->point.set1Value(n*j+i, t);
//       }
//     }
//    gridSep->addChild(coord);
// //  planeSep->addChild(texCoord);

    // add texture
    planeTexs[k] = new SoTexture2();
    planeTexs[k]->image.setValue(size, nc, black);
    planeTexs[k]->image.finishEditing();
    planeTexs[k]->model.setValue(SoTexture2::DECAL);
    planeTexs[k]->wrapS.setValue(SoTexture2::CLAMP);
    planeTexs[k]->wrapT.setValue(SoTexture2::CLAMP);
    gridSep->addChild(planeTexs[k]);

    // add mesh
    SoQuadMesh *quadMesh = new SoQuadMesh();
    quadMesh->verticesPerColumn.setValue(2);
    quadMesh->verticesPerRow.setValue(2);
    gridSep->addChild(quadMesh);

    planeSep->addChild(gridSep);

//     SoOutput *out = new SoOutput();
//     out->openFile(outFile);
//     SoWriteAction *wa = new SoWriteAction(out);
//     wa->apply(scene);
//     out->closeFile();
//     cerr << "Dumped scene graph to " << outFile << ".\n";
//     delete out;
//     delete wa;
     
  }


  // complexity
  SoComplexity *comp2 = new SoComplexity();
  comp2->value.setValue(0.05);
  scene->addChild(comp2);
  
  // add particles
//  SoSeparator *partSep = new SoSeparator();
  partSwitch = new SoSwitch();
//  partSwitch->addChild(partSep);
  partSwitch->whichChild.setValue(showParticles?SO_SWITCH_ALL:SO_SWITCH_NONE);
  scene->addChild(partSwitch);

  // add debug info
  partDebug = new SoSeparator();
  partSwitch->addChild(partDebug);

  // Set up keyboard
  SoEventCallback *callBack = new SoEventCallback();
  callBack->addEventCallback(SoKeyboardEvent::getClassTypeId(),
			     keyEventHandler);
  callBack->addEventCallback(SoEvent::getClassTypeId(),
			     mouseEventHandler, callBack);
  scene->addChild(callBack);


  // initialize simulation variables
  const int s = (dim[0]+2)*(dim[1]+2)*(dim[2]+2);
  //const int s = precomputeIdxs();
  u = new float[s];
  v = new float[s];
  w = new float[s];
  dens = new float[s];
  temp = new float[s];
  u0 = new float[s];
  v0 = new float[s];
  w0 = new float[s];
  du = new SbVec3f[s];
  dv = new SbVec3f[s];
  dw = new SbVec3f[s];
  dens0 = new float[s];
  temp0 = new float[s];
  vortg = new SbVec3f[s];
  //vortp = new SbVec3f[numParticles];
  
  resetSim();

  cerr << "Starting simulation.\n";


  // run simulation
  if (!renderQuit){

    cerr <<
      "\n"
      "Press ESC, then P to pause.\n"
      "\n"
      "Keys:\n"
      "  SPACE  pause simulation\n"
      "    D    write density to file\n"
      "    R    reset simulation\n"
      "    T    reset texture coordinates\n"
      "    W    write scene to file\n"
      "    P    toggle particles\n"
      "    M    toggle displaying force/speed/density/texture\n"
      "\n"
	 << endl;

    // This returns a main window to use.
    // If unsuccessful, exit.
    Widget myWindow = SoXt::init(argv[0]); // pass the app name
    if (myWindow == NULL) exit(1);

    // timer
    SoTimerSensor *sensor = new SoTimerSensor(animate, texCoord);
    sensor->setInterval(SbTime(0.033)); // 30 frames per second
    sensor->schedule();
    
    // Set up viewer
    SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
    myViewer->setTransparencyType(SoGLRenderAction::DELAYED_ADD);
    myViewer->setSceneGraph(scene);
    myViewer->viewAll();

    //myViewer->setCamera(camera);
    char title[20] = "";
    sprintf(title, "Fluid (e=%g)", vortEps);
    myViewer->setTitle(title);
    myViewer->show();
    
    SoXt::show(myWindow); // Display main window
    SoXt::mainLoop();	// Main Inventor event loop
  }
  else {
    while (t<maxTime){
      animate(NULL,NULL);
    }
  }
  
  return(0);
}



// cylinder
void showRay(SbVec3f pos, SbVec3f dir,
	     double height, SbColor color, int clear){
  static double hue = 0.0;
  double red, green, blue;

  // height stuff
  double cb_rat = 0.2;
  double cone_height = cb_rat*height;
  double bar_height = height-cone_height;
  double bar_radius = bar_height/10.0;
  double cone_radius = bar_height/3.0;

  // add self to scene graph
  static SoSeparator *showRaySep = NULL;
  if (showRaySep == NULL){
    // add cube sep to scene
    showRaySep = new SoSeparator();
    SoComplexity *comp = new SoComplexity();
    comp->value.setValue(0.3);
    showRaySep->addChild(comp);
    partDebug->addChild(showRaySep);
  }
  else if (clear){
    showRaySep->removeAllChildren();
    SoComplexity *comp = new SoComplexity();
    comp->value.setValue(0.3);
    showRaySep->addChild(comp);

    return;
  }
  
  // Ray debugging
  SoSeparator *sep = new SoSeparator();
  SoTransform *trn;
  SoMaterial *mat = new SoMaterial();
  mat->ambientColor.setValue(0.0, 0.0, 0.0);
  mat->diffuseColor.setValue(0.0, 0.0, 0.0);
  mat->specularColor.setValue(1.0, 1.0, 1.0);
  mat->shininess.setValue(0.3);
  mat->transparency.setValue(0);

  if (color[0]<0.0){
    hue += (1.0-hue)*0.05;
    mat->emissiveColor.setHSVValue(hue, 1.0, 0.8);
  }
  else {
    for (int i=0; i<3; i++)
      color[i] = 1.0-exp(-2.0*color[i]/10.0);
    mat->diffuseColor.setValue(color);
  }
  sep->addChild(mat);
    
  // Translate base
  trn = new SoTransform();
  trn->translation.setValue(pos);
  sep->addChild(trn);

  // Scale
  trn = new SoTransform();

  // Rotate
  SbRotation sbr;
  sbr.setValue(SbVec3f(0.0, 1.0, 0.0), dir);
  trn->rotation.setValue(sbr);
  sep->addChild(trn);

  // Pole
  SoCylinder *cyl = new SoCylinder();
  cyl->radius.setValue(bar_radius);
  cyl->height.setValue(bar_height);
  trn = new SoTransform();
  trn->translation.setValue(SbVec3f(0.0, bar_height/2.0, 0.0));
  sep->addChild(trn);
  sep->addChild(cyl);
  
  // Cone
  SoCone *con = new SoCone();
  con->bottomRadius.setValue(cone_radius);
  con->height.setValue(cone_height);
  trn = new SoTransform();
  trn->translation.setValue(SbVec3f(0.0, height/2.0, 0.0));
  sep->addChild(trn);
  sep->addChild(con);
  
  showRaySep->addChild(sep);
}
