//'WWWW'WWW''''''''WW''''''''WWW''WW''W'''''''''''''''''''''''''WWW'''''''
//WW  W  WW        WW         WW  WW  W                          WW       
//WW     WW   WW   WWW   WWW  WW  WW  W  WWW W WW  WWWW     WWW  WW   WWW 
//WW WW  WW  WW W  WW W W WW  WW  WW  W W WW WW   WWW      WW W  WW  W WW 
//WW  W  WW  WW W  WW W W WW  WW  WW W  W WW WW     WWW    WW    WW  W WW 
// WWWW WWWW  WW   WWW   WWW WWWW  WW    WWW WW   WWWW      WWW WWWW  WWW 
//                                                                        
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#ifndef GLOBALVARS
#define GLOBALVARS

//#include "Main.h"

#include <sys/time.h>
#include <sys/times.h>
#include <semaphore.h>
#include <string>

#include <Inventor/SbColor.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>

using namespace std;

class Pane;       // forward declaration of Pane class
class CameraList; // forward declaration of CameraList class
class PreviewQuad;
class SoSeparator;
class SoTransform;
class SoSwitch;
class SoCamera;

typedef struct timeRecord_st {
  int render;
  clock_t renderTicks;
  int totalTime;
  clock_t totalTicks;
  int octreeCreation;
  int photonShooting;
  int photonBalancing;
  int photonIrradiance;
} timeRecord_t;

class GlobalVars {
private:
public:
  enum tonemap_t {
    TM_NONE=0,
    TM_WARDCBSF,
    TM_WARDHIST,
    TM_WARDHUMAN,
    TM_NONLINEAR,
    TM_EXPOSURE
  };
  enum alg_t { ALG_RT=0, ALG_PM, ALG_MCPTEL, ALG_MCPT, ALG_FPS };
  enum accel_t { ACC_OCTREE=0, ACC_KDTREE };
  enum depthImage_t { DEPTH_NONE=0, DEPTH_NEAREST, DEPTH_AVERAGE };
  
  string sceneFile;
  string imgFile;
  string inImgFile;
  string ivOutFile;
  int imgDim[2];
  double imgMin[2], imgMax[2];
  int headLightPos;
  int renderQuit;
  int algorithm;
  int numSamples;
  int timeSamples;
  int maxDepth;
  int octreeDebug;
  int octreeMaxDepth;
  int octreeMaxItems;
  float luminanceMean;
  timeRecord_t times;
  int lightSamples;
  int lightSqrtSamples;
  int debugPixel[2];
  int debugRange;
  int animateScene;
  int areaLights;
  int automaticUpdate;
  int createBackup;
  int allowReflection;
  string cmdLine;
  double unitsPerMeter;
  double unitsPerMeter3;
  double metersPerUnit;
  double metersPerUnit2;
  double samples[100];
  SbColor windowBG;
  SbColor worldBG;
  SoSeparator *root;      // root of scenegraph
  SoSeparator *scene;     // scene that is rendered (child of root)
  SoTransform *trans;     // transform at head of scenegraph
  SoSeparator *letterSep; // heads-up display of letter key press
  SoSwitch *sceneSwitch;  // turns off scene
  SoSwitch *crossSwitch;  // turns off crosshair
  SoCamera *cam;          // points to Examiner viewer camera
  Pane *pane;             // points to pane rendering object
  CameraList *camList;    // points to CameraList for camera movement
  SoXtExaminerViewer *xtViewer;
  PreviewQuad *preview;
  int gPhotonEstimate;
  int cPhotonEstimate;
  int iPhotonEstimate;
  int vPhotonEstimate;
  float numPhotonsFireMax;
  int numPhotons;
  int numCPhotons;
  int numIPhotons;
  int numVPhotons;
  int approximate;
  double irradCacheError;
  int showPhotons;
  int toneMapping;
  int renderFlags;
  int gatherRays;
  int precomputeIrr;
  int fastTransparency;
  int numThreads;
  int nonRandomPixels;
  int quasiRandom;
  float ambientResolution;
  float refractionIndex;  // currently not used. See Material::ior2 instead.
  int tempInt;
  float tempFloat, tempFloat2, tempFloat3;
  int floatMap;
  float maxRelError;
  float timeVal;
  float timeMin, timeMax;
  int frames;
  int framesSegment[2];
  int rrStartDepth;
  int sceneHasShininess;
  int sceneHasLights;
  int seeSingleSided;
  int useGlassModel;
  int renderIv;
  int useDimensions;
  int figureMode;
  int showNormals;
  float gamma;
  float exposure;
  string comment;
  float fNumber;
  float transpScale;
  string environMapFile;
  int envLockSamples;
  float envMapRotation;
  string sampleMap;
  int silent;
  int useEmisFluxDensity;
  int computeNormals;
  int useGlare;
  int srandSeed;
  int simpleOctree;
  int prepass;
  int makeMatte;
  SbColor attenuation;
  int useIndirectImpSampling;
  int bumpMapping;
  float bumpScale;
  int useIrradCache;
  float heightAngle; // useful for tone mapping
  float sigma_t;
  float albedo;
  float stepSize;
  string scalarDataFile;
  string scalarDataFile2;
  int scalarDataDim[3];
  int scalarDataHeader;
  float scalarDataMult;
  float scalarDataMult2;
  SbVec3f scalarDataMax;
  SbColor scalarColor;
  int localIllum;
  int ambientIllum;
  accel_t accelType;
  depthImage_t depthImage;
  float compressEst;
  bool useProjectionMaps;
  int addPlane;
  int addDMPlane;
  string waterMark;
  
  // thread stuff
  static const int maxThreads=32;
  pthread_t threads[maxThreads];
  sem_t glob_sema[maxThreads];

  GlobalVars();
  ~GlobalVars();

  void readOpts(int argc, char *argv[]);
  void printUsage(int version=0);

  double setUnitsPerMeter(double upm){
    unitsPerMeter = upm;
    unitsPerMeter3 = upm*upm*upm;
    metersPerUnit = 1.0/upm;
    metersPerUnit2 = 1.0/(upm*upm);
  }
  
  double meters(double units){  return(units*metersPerUnit);  }
  double meters2(double units){ return(units*metersPerUnit2); }
  double units(double meters){  return(meters*unitsPerMeter); }
  double units3(double meters){ return(meters*unitsPerMeter3); }
};

#endif /* GLOBALVARS */
