//WW''''WW'''''''WW''''''''''''WW''''''WWW''''''''''WWW''''''''''''''''''
//WWW  WWW       WW                     WW           WW                  
//W WWW WW  WWW WWWW  WW  W WW WW  WWW  WW      WWW  WW   WWW  WWWW  WWWW
//W  W  WW W WW  WW  WWWW WW   WW W WW  WW     WW W  WW  W WW WWW   WWW  
//W     WW W WW  WW  WW   WW   WW W WW  WW     WW    WW  W WW   WWW   WWW
//W     WW  WWW  WW   WWW WW   WW  WWW WWWW     WWW WWWW  WWW WWWW  WWWW 
//                                                                       
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#include <assert.h>
#include <Inventor/nodes/SoMaterial.h>
#include "Material.h"
#include "SoMatPlus.h"
#include "debug.h"
#include "Main.h"
#include "Misc.h"

using namespace std;

Material::Material(){
  init();
}

Material::Material(SoMaterial *mat, SoMatPlus *matPlus) {
  init(mat, matPlus);
}

void Material::init(SoMaterial *mat, SoMatPlus *matPlus) {
  static int newMats = 0;

  bool delMat = false;
  bool delMatPlus = false;
  if (!mat) {
    mat = new SoMaterial();
    mat->ref();
    delMat = true;
  }
  if (!matPlus) {
    matPlus = new SoMatPlus();
    matPlus->ref();
    delMatPlus = true;
  }
  
  ambient      = mat->ambientColor[0];
  diffuse      = mat->diffuseColor[0];
  setShininess(mat->shininess[0]);
  setGlossiness(0);
  setSpecular(mat->specularColor[0]);
  setEmission(mat->emissiveColor[0]);
  transparency = mat->transparency[0];
  
  // mat plus
  model     = (ReflModel) matPlus->model.getValue();
  rough1    = matPlus->rough1.getValue();
  rough2    = matPlus->rough2.getValue();
  isotropy1 = matPlus->isotropy1.getValue();
  isotropy2 = matPlus->isotropy2.getValue();
  R0        = matPlus->R0.getValue();
  layers    = matPlus->layers.getValue();
  ior1      = matPlus->ior1.getValue();
  ior2      = matPlus->ior2.getValue();
  atten     = matPlus->atten.getValue();
  emisEcc   = matPlus->emisEcc.getValue();
  minDist   = matPlus->minDist.getValue();

  resolveParams();
  
  for (int j=0; j<3; j++)
    if (!finite(diffuse[j]))
      diffuse = SbColor(1.0, 0.0, 0.0);

  // temporary reflection capping
  const static bool reflCap = (bool)getenv("PANE_REFLECTION_CAP");
  if (reflCap) diffuse = BBox::vecMin(diffuse, SbVec3f(.9,.9,.9));

  const static bool reflScale = (bool)getenv("PANE_REFLECTION_SCALE");
  if (reflScale)
    diffuse *= .8;
  
  if (delMat) mat->unref();
  if (delMatPlus) matPlus->unref();
  
  //cerr << "Material():" << *this;
}

Material::~Material(){
}

void Material::resolveParams(){
  
  if (model == SCHLICK) {   // Override parameters if using Schlick

    // Schlick model (single layer) only has one color
    // so just use diffuse.
    specular = diffuse;
    
    float g = 0, s = 0, d = 0;
    g = 4*rough1*(1-rough1);
    
    // compute g, s, d
    if (rough1<0.5){
      s = 1-g;
      d = 0;
    } else {
      s = 0;
      d = 1-g;
    }

    // for debugging only
//     s = d = 0;
//     g = 1;

    shininess = s;
    glossiness = g;
    diffuseness = d;
  } else if (model == ASHIKHMIN){

    // scale diffuse by (1-R_s) now
    vecScaleItself(diffuse, SbVec3f(1,1,1)-specular);

//     // compute splitting probabilities
//     float specAvg = vecAvg(specular);
//     float diffAvg = vecAvg(diffuse);
//     float total = specAvg + diffAvg;

//     // smooth
//     specAvg += .25*total;
//     diffAvg += .25*total;
//     total += .5*total;
    
//     shininess = 0;
//     glossiness = specAvg/total;  // dont forget to divide by these, because each
//     diffuseness = diffAvg/total; // should be 1!

    // Even weights seems to produce a lot less noise for
    // bunny on plane scene with:
    // bunny: Rs=.6  Rd=0  nu=nv=100
    // plane: Rs=.05 Rd=.5 nu=nv=40000
    shininess = 0;
    glossiness = .5;
    diffuseness = .5;

    if (vecAvg(diffuse) <= 0){
      glossiness = 1;
      diffuseness = 0;
    }
  }
}

ostream& operator<< (ostream& os, const Material& m){
  os << "Material [" << "\n"
     << "  ambient      = " << m.ambient << "\n"
     << "  diffuse      = " << m.diffuse << "\n"
     << "  specular     = " << m.specular << "\n"
     << "  emission     = " << m.emission << "\n"
     << "  emissive     = " << (m.emissive?"yes":"no") << "\n"
     << "  shininess    = " << m.shininess << "\n"
     << "  transparency = " << m.transparency << "\n"
     << "  emisAvg      = " << m.emisAvg << "\n"
     << "  model        = " << m.model << "\n"
     << "  rough1       = " << m.rough1 << "\n"
     << "  rough2       = " << m.rough2 << "\n"
     << "  isotropy1    = " << m.isotropy1 << "\n"
     << "  isotropy2    = " << m.isotropy2 << "\n"
     << "  R0           = " << m.R0 << "\n"
     << "  layers       = " << m.layers << "\n"
     << "  ior1         = " << m.ior1 << "\n"
     << "  ior2         = " << m.ior2 << "\n"
     << "  atten        = " << m.atten << "\n"
     << "  emisEcc      = " << m.emisEcc << "\n"
     << "  minDist      = " << m.minDist << "\n"
     << "]\n";
  
  return os;
}

