#include <iostream>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoIndexedShape.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoGroup.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <Inventor/nodes/SoCoordinate3.h>

#include "Misc.h"

using namespace std;


//''''''''''''''''''''''''''''''WW'''''''WW'''W'''''''''''''''''''''''''WWW''
//                              WW       WWW  W                          WW  
// WWW  WW   WWWW W  WWW  W WW WWWW  WW  W WW W  WW   W WW WWWW W   WWW  WW  
//WW W WW W  WW WW W WW W W WW  WW  WWWW W  WWW WW W  WW   WW WW W W WW  WW  
//WW   WW W  WW WW W WW W W WW  WW  WW   W   WW WW W  WW   WW WW W W WW  WW  
// WWW  WW   WW WW W WWW   WWW  WW   WWW W    W  WW   WW   WW WW W  WWW WWWW 
//                   WW                                                      
//'''''''''''''''''''WW''''''''''''''''''''''''''''''''''''''''''''''''''''''
 
int computeNormals(SoNode *scene){

  int totalTris = 0;
  int debugLimit = 1000;
  int totalComputedNormals = 0;
  
  // search for shapehints
  SoSearchAction hintsSearch;
  hintsSearch.setType(SoShapeHints::getClassTypeId(), TRUE);
  hintsSearch.setSearchingAll(FALSE); // don't look under off switches
  hintsSearch.setFind(SoSearchAction::TYPE);
  hintsSearch.setInterest(SoSearchAction::LAST);
  
  // search for indexed shape
  SoSearchAction sa;
  sa.setType(SoIndexedShape::getClassTypeId(), TRUE);
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.setFind(SoSearchAction::TYPE);
  sa.setInterest(SoSearchAction::ALL);
  sa.apply(scene);
  SoPathList plist = sa.getPaths();
  int len = plist.getLength();
  cerr << len << " shape paths found\n";

  //WW''''''''''''''''''''''''''''''''''''''''''W''''''''''''''''''''
  // W                                          W                    
  // W   W   W  WW      W  W W  WW   WW     WWW WW   WW WW   WW   WWW
  // W  W W W W W W    W W W W Wmmm W       WW  W W W W W W Wmmm  WW 
  //WWW  W   W  WW      W   W   WW  W      WWW  W W  WW WW   WW  WWW 
  //''''''''''''W'''''''''''''''''''''''''''''''''''''''W''''''''''''
  for (int i=0; i<len; i++){

    // get IndexedShape
    SoPath *path = (SoPath*) plist[i];
    SoIndexedShape *shape = (SoIndexedShape *)path->getTail();
    cerr << "indexed shape ";

    // find vertex and normal lists
    SoMFVec3f *verts = NULL;
    SoMFVec3f *norms = NULL;
        
    if (shape->vertexProperty.getValue() == NULL){

      // search for Coordinate3 node
      cerr << "Searching for Coordinate3 node...";
      SoSearchAction sa2;
      sa2.setType(SoCoordinate3::getClassTypeId());
      sa2.setSearchingAll(FALSE); // don't look under off switches
      sa2.setInterest(SoSearchAction::LAST);
      sa2.apply(path);
    
      SoPath *coordPath = sa2.getPath()->copy();
      if (coordPath == NULL){
	cerr << "Could not find Coordinate3 node!\n";
	continue;
      }
      else
	cerr << "found.\n";
      
      SoCoordinate3 *coord = (SoCoordinate3 *)coordPath->getTail();
      verts = &(coord->point);

      // search for Normal node
      cerr << "Searching for Normal node...";
      sa2.setType(SoNormal::getClassTypeId());
      sa2.apply(path);
    
      SoPath *normPath = sa2.getPath();
      SoNormal *norm = NULL;
      
      if (normPath == NULL){
	
	cerr << "Could not find Normal node!\n";
	
	// create normal node
	norm = new SoNormal();
	SoGroup *coordParent = (SoGroup*) coordPath->getNodeFromTail(1);
	int coordIdx = coordParent->findChild(coord);
	coordParent->insertChild(norm, coordIdx+1);
	
      }
      else {
	cerr << "found.\n";
	norm = (SoNormal *)normPath->getTail();
      }
      
      norms = &(norm->vector);

    }
    else {
      
      // get VertexProperty
      SoVertexProperty *vprop = (SoVertexProperty *)shape->vertexProperty.getValue();
      vprop->normalBinding.setValue(SoVertexProperty::PER_VERTEX_INDEXED);
      shape->normalIndex.setValue(-1);
      verts = &(vprop->vertex);
      norms = &(vprop->normal);
    }

    // search for ShapeHints
    int reverseOrdering = 0;
    hintsSearch.apply(path);
    if (hintsSearch.getPath()){
      SoShapeHints *shapeHints = (SoShapeHints *)((SoFullPath *)hintsSearch.getPath())->getTail();

      // reverse vertex ordering based on ShapeHints
      if (shapeHints->vertexOrdering.getValue() == SoShapeHints::CLOCKWISE)
	reverseOrdering = 1;
    }
        
    // get number of vertices
    cerr << "\n";
    int numVerts = verts->getNum();
    int numCoordIdxs = shape->coordIndex.getNum();
    
    //W'''''''''''''''''''''''''''''''''''''''''''''''m'''''''''''''WW'''''''''''
    //W                                        W                     W           
    //W     W   W  WW      W  W W  WW   WW    WWW  WW W  WW WW   WW  W   WW   WWW
    //W    W W W W W W    W W W W Wmmm W       W  W   W W W W W W W  W  Wmmm  WW 
    //WWWW  W   W  WW      W   W   WW  W       W  W   W  WW W W  WW WWW  WW  WWW 
    //             W                                              W              
    //'''''''''''''W''''''''''''''''''''''''''''''''''''''''''''WW'''''''''''''''

    // data structures
    float *normalWeights = new float[numVerts];
    for(int j=0; j<numVerts; j++)
      normalWeights[j] = 0.0;

    // variables
    int v[3];
    v[0] = v[1] = v[2] = -1;
    int triCount = 0;
    int numTris = 0;
    int smallAreas = 0;
    int colinearVerts = 0;

    // loop over indices
    for (int j=0; j<numCoordIdxs; j++){

      // proces groups of 3 indices
      v[0] = v[1];
      v[1] = v[2];
      v[2] = shape->coordIndex[j];
      if (v[0] >= 0 && v[1] >= 0 && v[2] >= 0){

	// triangle found
	numTris++;             // for counting triangles
	triCount++;            // for flipping odd triangles
	if (numTris%1000==0)
	  printf("\r%6.2f%%    ", j/(float)(numCoordIdxs)*100.0);

	// detect same vertices
	if (v[0] == v[1] || v[0] == v[2] || v[1] == v[2]){
	  if (colinearVerts++ == 0)
	    cerr << "Warning! Colinear vertices in triangle!\n";
	  continue;
	}
	
	// get center
	SbVec3f vert[3];
	vert[0] = (*verts)[v[0]];
	vert[1] = (*verts)[v[1]];
	vert[2] = (*verts)[v[2]];
	
	// get normal
	SbVec3f normal;
	SbVec3f edge1 = vert[1] - vert[0];
	SbVec3f edge2 = vert[2] - vert[0];
	
	if (reverseOrdering == 0)
	  normal = edge1.cross(edge2);
	else
	  normal = edge2.cross(edge1);
	
	// flip odd triangles if IndexedTriangleStripSet
	if (shape->isOfType(SoIndexedTriangleStripSet::getClassTypeId()) &&
	    triCount%2==0)
	  normal = -normal;

	float area2 = normal.normalize();  

	if ((area2/vert[0].length())<0.00000001){
	  printf("\r%6.2f%%    ", j/(float)(numCoordIdxs)*100.0);
	  cerr << smallAreas++ << " areas too small, skipping.";
	  continue;
	}

	float area = area2/2.0;
	
	// add tri to vert's tri lists
	for (int k=0; k<3; k++){

	  // initialize
	  if (normalWeights[v[k]] <= 0.0)
	    norms->set1Value(v[k], SbVec3f(0.0, 0.0, 0.0));

	  norms->set1Value(v[k], (*norms)[v[k]] + normal*area);
	  normalWeights[v[k]] += 1.0f/area;
	}
	
      }
      else
	triCount = 0;
      
    } // end coords/triangle loop
    printf("\r%6.2f%%                                                   \n", 100.0);
    totalTris += numTris;

    //''''''''''''''''''''''''''''''''''''''''''''''''''''''''WW''''''
    //                       W                                 W      
    // WW  W  WW W  WW  W W WWW  WW     WW   W   WW WW W   WW  W   WWW
    //W   W W W W W W W W W  W  Wmmm    W W W W W   W W W W W  W   WW 
    // WW  W  W W W WW   WW  W   WW     W W  W  W   W W W  WW WWW WWW 
    //              W                                                 
    //              W                                                 
    //''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    int computedNormals = 0;

    // for each vert
    int badNormals = 0;
    for(int j=0; j<numVerts; j++){

      // normalize normal
      if (normalWeights[j]>0.0){
	SbVec3f normal = (*norms)[j]/normalWeights[j];
	normal.normalize();
	norms->set1Value(j, normal);
 	computedNormals++;
      }
      else
	badNormals++;
    }
    totalComputedNormals += computedNormals;
    cerr << computedNormals << " vertex normals computed ("
	 << badNormals << " bad)" << endl;
    delete [] normalWeights;
  }

  return(totalComputedNormals);
}





