
#include <stdlib.h>
#include <iostream>
#include "Noise.h"
#include "debug.h"

// #define float double
// #define powf pow
// #define logf log

using namespace std;

void Noise::init(){

  // Gradient directions from:
  // "Improved Noise" by Ken Perlin
  G[0]  = SbVec3f( 1, 1, 0);
  G[1]  = SbVec3f(-1, 1, 0);
  G[2]  = SbVec3f( 1,-1, 0);
  G[3]  = SbVec3f(-1,-1, 0);

  G[4]  = SbVec3f( 1, 0, 1);
  G[5]  = SbVec3f(-1, 0, 1);
  G[6]  = SbVec3f( 1, 0,-1);
  G[7]  = SbVec3f(-1, 0,-1);

  G[8]  = SbVec3f( 0, 1, 1);
  G[9]  = SbVec3f( 0,-1, 1);
  G[10] = SbVec3f( 0, 1,-1);
  G[11] = SbVec3f( 0,-1,-1);

  G[12] = SbVec3f( 1, 1, 0);
  G[13] = SbVec3f(-1, 1, 0);
  G[14] = SbVec3f( 0,-1, 1);
  G[15] = SbVec3f( 0,-1,-1);
  
  // compute array of random directions G
  for (int i=0; i<n; i++){
    G[i].normalize();
  }

  // compute permuted set of integers, P

  // initialize ordered list of integers
  for (int i=0; i<n; i++)
    P[i] = i;

  RNG rng;
  seed = 15;
  rng.setSeed(seed);
  
  // shuffle order
  for (int i=n-1; i>0; i--){
      
    int j = (int)(rng()*i);

    // swap i and j
    int t = P[i];
    P[i] = P[j];
    P[j] = t;
  }


  sign = true;
  amplitude = 1;
  bias = 0.5;
  gain = 0.5;
  translation = SbVec3f(0,0,0);
  frequency = SbVec3f(1,1,1);
  offset = 0;
  ampJump = 0.5;
  freqJump = 2;
  octaves = 1;
  //funcPtr = &Noise::monofractal;
  //funcPtr = &Noise::multiFractal;
  //funcPtr = &Noise::hybridMultiFractal;
  //funcPtr = &Noise::fBm;
  funcPtr = &Noise::ridgedMultiFractal;
}

float Noise::perlin(const SbVec3f &x_) const {

  // FPE if: INT_MAX < x[i] < INT_MIN 

  SbVec3f x = x_;
  int xi0, yi0, zi0;

//   const int INT_LRG = INT_MAX/2;
//   const int INT_SML = -INT_LRG;

//   if (x[0]>INT_LRG) x[0] = INT_LRG;
//   else if (x[0]<INT_SML) x[0] = INT_SML;

//   if (x[1]>INT_LRG) x[1] = INT_LRG;
//   else if (x[1]<INT_SML) x[1] = INT_SML;
  
//   if (x[2]>INT_LRG) x[2] = INT_LRG;
//   else if (x[2]<INT_SML) x[2] = INT_SML;

  xi0 = (int)floor(x[0]);
  yi0 = (int)floor(x[1]);
  zi0 = (int)floor(x[2]);

  int xi1, yi1, zi1;
  xi1 = xi0+1;
  yi1 = yi0+1;
  zi1 = zi0+1;

  float x0, y0, z0;
  x0 = x[0]-xi0;
  y0 = x[1]-yi0;
  z0 = x[2]-zi0;

  float x1, y1, z1;
  x1 = x[0]-xi1;
  y1 = x[1]-yi1;
  z1 = x[2]-zi1;

  float omegax0 = omega(x0);
  float omegay0 = omega(y0);
  float omegaz0 = omega(z0);

  float omegax1 = 1-omegax0;//omega(x1);
  float omegay1 = 1-omegay0;//omega(y1);
  float omegaz1 = 1-omegaz0;//omega(z1);
  
  return
    TauDotXYZ(xi0, yi0, zi0, x0, y0, z0)*omegax0*omegay0*omegaz0 +
    TauDotXYZ(xi0, yi0, zi1, x0, y0, z1)*omegax0*omegay0*omegaz1 +
    TauDotXYZ(xi0, yi1, zi0, x0, y1, z0)*omegax0*omegay1*omegaz0 +
    TauDotXYZ(xi0, yi1, zi1, x0, y1, z1)*omegax0*omegay1*omegaz1 +
    TauDotXYZ(xi1, yi0, zi0, x1, y0, z0)*omegax1*omegay0*omegaz0 +
    TauDotXYZ(xi1, yi0, zi1, x1, y0, z1)*omegax1*omegay0*omegaz1 +
    TauDotXYZ(xi1, yi1, zi0, x1, y1, z0)*omegax1*omegay1*omegaz0 +
    TauDotXYZ(xi1, yi1, zi1, x1, y1, z1)*omegax1*omegay1*omegaz1;
}


float Noise::monofractal(const SbVec3f &x_) const {
  SbVec3f x = vecScale(x_ + translation, frequency);
  float amp = 1;
  float freq = 1;
  float val = 0;
  for (int j=0; j<octaves; j++){
    val += amp*perlin(x*freq);
    amp *= ampJump;
    freq *= freqJump;
  }

  return applyRest(val);
}

// from Texturing & Modeling: A Procedural Approach pg. 440
float Noise::fBm(const SbVec3f &x_) const {
  SbVec3f x = vecScale(x_ + translation, frequency);
    
  float mfoffset = 0; //.8;
  float val = 0.0;

  for (int i=0; i<octaves; i++){
    val += (perlin(x)+mfoffset) * powf(freqJump, -ampJump*i);
    x *= freqJump;
  }
  return applyRest(val);
}

// Methods for Realistic Lanscape Imaging by Ken Musgrave
// Terrain Model p. 55
float Noise::terrain(const SbVec3f &x_) const {
    
  SbVec3f x = vecScale(x_ + translation, frequency);
  float amp = 1;
  float freq = 1;
  float val = 0;
  for (int j=0; j<octaves; j++){
    val += amp*perlin(x*freq);
    amp *= ampJump;
    freq *= freqJump;
  }

  return applyRest(val);
}
  
// from Texturing & Modeling: A Procedural Approach pg. 440
float Noise::multiFractal(const SbVec3f &x_) const {
  SbVec3f x = vecScale(x_ + translation, frequency);
    
  float mfoffset = 0.8;
  float val = 1.0;
  float tot = 0;

  for (int i=0; i<octaves; i++){
    val *= (perlin(x)+mfoffset) * powf(freqJump, -ampJump*i);
//      tot += val;
    x *= freqJump;
  }
  //cerr << "val=" << val << " tot=" << tot << "\n";
  return applyRest(val);
  //return applyRest(tot);
}

// from Texturing & Modeling: A Procedural Approach pg. 502
float Noise::hybridMultiFractal(const SbVec3f &x_) const {

  // precompute exponents
  static vector<float> exponentArray(octaves+1);
  static float lastAmpJump, lastFreqJump;
  if (exponentArray.empty() ||
      lastAmpJump != ampJump || lastFreqJump != freqJump) {
    float freq=1;
    for (int i=0; i<=octaves; i++){
      exponentArray[i] = powf(freq, -ampJump);  // ampJump = "H"
      freq *= freqJump;
    }
  }

  SbVec3f x = vecScale(x_ + translation, frequency);
  float hyoffset = 0.8;
  float val = (perlin(x) + hyoffset)*exponentArray[0]; // "result"
  float weight = val;

  for (int i=1; i<octaves; i++){

    // prevent divergence
    if (weight>1) weight = 1;

    float signal = (perlin(x) + hyoffset)*exponentArray[i];
    val += weight*signal;

    weight *= signal;
    x *= freqJump;
  }

  return applyRest(val);
}

// from Texturing & Modeling: A Procedural Approach pg. 504
float Noise::ridgedMultiFractal(const SbVec3f &x_) const {

  // precompute exponents
  //
  // WARNING: THIS IS NOT THREAD SAFE!
  const int maxOctaves=30;
  static vector<float> exponentArray(maxOctaves+1);
  static float lastAmpJump, lastFreqJump;
  if (exponentArray.empty() ||
      lastAmpJump != ampJump || lastFreqJump != freqJump) {
    float freq=1;
    for (int i=0; i<=maxOctaves; i++){
      exponentArray[i] = pow(freq, -ampJump);  // ampJump = "H"
      freq *= freqJump;
    }
    lastAmpJump = ampJump;
    lastFreqJump = freqJump;
  }

  SbVec3f x = vecScale(x_ + translation, frequency);
  float signal = perlin(x);
  if (signal<0) signal = -signal;
  float rmoffset = 1.0;
  signal = rmoffset-signal;
  signal *= signal;
  float val = signal;
  float weight = 1.0;

  for (int i=1; i<octaves; i++){

    x *= freqJump;
    weight = signal*gain;
    // prevent divergence
    if (weight>1) weight = 1;
    if (weight<0) weight = 0;
    signal = perlin(x);
    if (signal<0) signal = -signal;
    float rmoffset = 1.0;
    signal = rmoffset-signal;
    signal *= signal;

    signal *= weight;
    val += signal*exponentArray[i];
  }

  //return applyRest(val);
  // applyRest w/o gain...
  if (!sign){
    val = fabs(val);
    val = biasFunc(val, bias);
  } else if (val>=0) {
    val = biasFunc(val, bias);
  } else  {
    val = -biasFunc(-val, bias);
  };
    
  return val*amplitude+offset;
  
} 


















//g++ Noise.cxx -I/usr/X11R6/include -I/usr/include/OpenEXR -lInventor -O2 -DNOISE_MAIN -o noiseTest
#ifdef NOISE_MAIN
int main(int argc, char *argv[]){
#else
int NoiseMain(int argc, char *argv[]){
#endif

  if (argc != 2){
    cerr << "Usage: " << argv[0] << " n \n";
    return 1;
  }

  Noise myNoise;
  
  //srand48(15);
  int n = atoi(argv[1]);
  float invN = 100./n;
  float total=0;
  for (int k=0; k<n; k++)
    for (int j=0; j<n; j++)
      for (int i=0; i<n; i++){
        total += myNoise.noise(SbVec3f(i*invN, j*invN, k*invN));
      }

  cerr << "Result = " << total << endl;
  
  return 0;
}
