#include "debug.h"
#include "PixelIterator.h"

// syntax: while (iter.next(x,y)){ ... }
// we're combining initializing and iterating/testing, which were usually done
// before/after loop. this alters the order.
// 1) initialize in ctor.
// 2) iterate, but only test/return old values
bool PixelIterator::next(int &x, int &y){
  if (started) goto resume;
  started=true;
    
  for (j=lowerLeft[Y]+orderId; j<upperRight[Y]; j+=numThreads){

    // completion meter
    completion[id] = j*100.0/imgDim[Y];
    printStatus();

    for (i=lowerLeft[X]; i<upperRight[X]; i++){

      x=i;
      y=j;
      return true; // render pixel

      // re-enter here
      resume:
      1; 
      
    }
  }
  
  return false; // done
}

void PixelIterator::printStatus(float val){
  if (val>=0) completion[i]=val;
  if (id!=0) return;

  // print every second
  int lastSec = lastTime.tv_sec;
  gettimeofday(&lastTime, NULL);
  if (lastTime.tv_sec == lastSec) return;

  pthread_mutex_lock(&mutexes[LOCK_STDERR]);
  print("\r");
  for (int i=0; i<numThreads; i++)
    print("%6.2f%%%c ", completion[i], completion_stage[i]);
  pthread_mutex_unlock(&mutexes[LOCK_STDERR]);
  
};

bool ProgressiveIterator::next(int &x, int &y){
  if (started) goto resume;
  started=true;

  // loop over shrinking tile sizes
  while (tileSize>=2){

    // loop over shrinking tiles covering image
    for (tileJ=0; tileJ*tileSize<(upperRight[Y]-lowerLeft[Y]); tileJ++){

      // completion meter
      completion[id] = j*100.0/imgDim[Y];
      printStatus();
	
      for (tileI=0; tileI*tileSize<(upperRight[X]-lowerLeft[X]); tileI++){

        // each tile has 2x2 subtiles
        for (innerJ=0; innerJ<2; innerJ++){
          for (innerI=0; innerI<2; innerI++){

            // handle special case of first pixel/tile (entire image)
            if (!firstDrawn)
              firstDrawn = 1;
            // skip tiles bottom left cell (it was computed in previous pass)
            else if (innerI==0 && innerJ==0) 
              continue;

            i = lowerLeft[X]+tileSize*tileI+innerI*tileHalfSize;
            j = lowerLeft[Y]+tileSize*tileJ+innerJ*tileHalfSize;

            // only render within render area
            if (i>=upperRight[X] || j>=upperRight[Y])
              continue;

            // only render pixels assigned to this thread
            if (pixelCount++%numThreads != orderId)
              continue;

            // render pixel (i,j)
            x=i;
            y=j;
            return true;

            // re-enter here
            resume:
            1;
            
          } // innerI
        } // innerJ

        
      } // tileI
    } // tileJ

    // refine tile size
    tileSize = tileHalfSize;
    tileHalfSize = tileHalfSize >> 1;
    tileIteration++;
    
  } // tileSize

  return false;
}


void ProgressiveIterator::printStatus(float val){
  if (val>=0) completion[i]=val;
  if (id!=0) return;

  // print every second
  int lastSec = lastTime.tv_sec;
  gettimeofday(&lastTime, NULL);
  if (lastTime.tv_sec == lastSec) return;

  pthread_mutex_lock(&mutexes[LOCK_STDERR]);
  print("\rtileSize: %d (>=2)" , tileSize);
  for (int i=0; i<numThreads; i++)
    print("%6.2f%%%c ", completion[i], completion_stage[i]);
  print("      ");
  pthread_mutex_unlock(&mutexes[LOCK_STDERR]);
};
