//WWWW''''''''''''''''''''WWW''''''''''''''''''
//WW  W                    WW                  
//WW  W  WWW W WW     WWW  WW   WWW  WWWW  WWWW
//WWWW  W WW W WW    WW W  WW  W WW WWW   WWW  
//WW  W W WW W WW    WW    WW  W WW   WWW   WWW
//WW  W  WWW  WWW     WWW WWWW  WWW WWWW  WWWW 
//             WW                              
//'''''''''''WWW'''''''''''''''''''''''''''''''

#ifndef RAYCLASS
#define RAYCLASS

#include <iostream>
#include <Inventor/SbLinear.h>
#include <Inventor/SbColor.h>

// for piggyback variables
class Triangle;  
class LineSet;
class Primitive;
typedef struct thread_context_st thread_context;

using namespace std;

class Ray {
private:

  static LineSet *ls;
  
public:

  static Primitive *NO_PRIM;
  
  SbVec3f pos;
  SbVec3f dir;             // must be normal (unit length)

  int depth;               // depth from eye of hit point (first is 1)

  int atten;               // Used by *dielectric* media to track state.
                           // True if in inside dielectric and Beer's law
                           // should be applied. Used by getRadiance1, 2,
                           // emitPhotons.
  SbColor attenColor;      // The attenuation to use.
  
  int inside;              // Used by *volumetric* media to track state.
                           // Used by volumeIntegrate to not have density
                           // inside glass.
  
  float time;              // time of ray
  
  float importance;        // importance of result of ray
                           // (1.0 == very important, primary)
                           // Used by refractionContrib and getRadiance1.

  int ldInstance;          // instance i
  int ldDimensionThis;     // Low discrepancy sequence dimension for this ray.
                           // Starts at 0. Identifies the first available dimension.
  int ldDimensionNext;     // First available low discrepancy sequence
                           // dimension for CHILD rays, with 0-based starting index.
  
  // piggy back variables for intersection
  thread_context *ctx;     // context for access to thread global vars/RNG's

  // input
  Primitive *ignorePrim;            // id of triangle ray should not hit, if >=0
  
  // early output
  //
  // These should be set by intersectRay
  //
  // No hit:
  //   intersectRay returns false
  //   t = -1
  //   tri = NULL
  //   XXX triId = Ray::NO_TRI          (triId not currently used)
  //   bc = UNDEFINED
  //
  // Hit:
  //   intersectRay returns true
  //   t = dist to closest intersection
  //   tri = Primitive* of primitive intersected
  //   XXX triId = id of primitive      (triId not currently used)
  //   bc = bary of prim
  //
  double t;                // length along ray to hit point 
  double tMedia;           // length until first participating media scatter.
                           // Used to detect light contribution,
                           // by multipleDiffuseContrib for irradianceCaching
  Primitive *prim;         // pointer to primitive that this ray hits
  SbVec3f bc;              // barycentric coordinates of hit point

  // eventual output
  SbVec3f x;               // position in space of the hit point
  SbColor diffCol;         // diffuse color of hit point surface

  
  Ray() : pos(0,0,0), dir(0,0,-1), importance(1), atten(0), inside(0),
          ignorePrim(NO_PRIM) { }
  
  Ray(const SbVec3f & pos_, const SbVec3f & dir_, thread_context *ctx_,
      Primitive *ignorePrim_=NO_PRIM, int depth_=1):
    pos(pos_), dir(dir_), ctx(ctx_), depth(depth_),
    importance(1),
    atten(0),
    inside(0),
    ignorePrim(ignorePrim_) { }
  
  Ray(const SbVec3f &pos_, const SbVec3f &dir_, const Ray &r,
      Primitive *ignorePrim_=NO_PRIM) :
    pos(pos_), dir(dir_),
    ctx(r.ctx),
    depth(r.depth+1),                        // depth auto. incremented
    importance(r.importance),
    atten(r.atten),
    attenColor(r.attenColor),
    inside(r.inside),
    ignorePrim(ignorePrim_),
    ldInstance(r.ldInstance),
    ldDimensionThis(r.ldDimensionNext),      // ldDimension auto. incremented 
    ldDimensionNext(r.ldDimensionNext),
    prim(NULL),
    time(r.time) { }
  
  // returns point that is parameter t along the ray
  SbVec3f point(float t){ return pos + t*dir; }
  SbVec3f point(){ return pos + t*dir; }

  void ignore(Primitive *p){ ignorePrim = p; }

  // draws cylinder representation
  void show(double height = 100.0, SbColor color = SbColor(-1.0, -1.0, -1.0),
	    int clear = 0);

  
  // draws line representation
  void show2(double height = 100.0, SbColor color = SbColor(-1.0, -1.0, -1.0),
	     int clear = 0);

  friend ostream& operator<< (ostream& os, const Ray& r);
};

#endif /* RAYCLASS */
