
#ifndef RAYACCELCLASS
#define RAYACCELCLASS

#include "Primitive.h"

class Ray; // forward declaration of Ray class

class RayAccel : public Primitive {
  
private:

  void error(const char *f);
  
protected:
  int numPrims;
  double maxSideLength;
  SbVec3f min, max;

  // keep bounding sphere of primitives
  SbVec3f center;
  float radius;
  float shiny_radius;
  SbVec3f shiny_center;
  
public:

  RayAccel(){
    //error("RayAccel");
    numPrims = 0;
    maxSideLength = 0;
    min[0] = min[1] = min[2] = 0;
    max[0] = max[1] = max[2] = 0;
    center.setValue(0,0,0);
    radius = shiny_radius = 0;
    shiny_center.setValue(0,0,0);
  }
  virtual ~RayAccel(){ }
  virtual void reset(){ error("reset"); }
  virtual int add(Primitive *t){ error("add"); return 0; }
  virtual void reserve(int n) { error("reserve"); }
  virtual void printStats(){ error("printStats"); }
  virtual void setParams(int maxD, int maxI, int debug){ error("setParams"); }
  virtual void refine(int simpleDiv = 0){ error("refine"); }
  virtual void reorderNodes(){ cerr << "error: RayAccel::reorderNodes()!\n"; }

  // Primitive methods
  SbVec3f getMin() const { return min; }
  SbVec3f getMax() const { return max; }
  
  int size(){ /*Error("size");*/ return(numPrims); }
  double& getMaxSideLength(){ /*Error("getMaxSideLength");*/ return maxSideLength; }
  SbVec3f& getCenter(){ /*Error("getCenter");*/ return center; }
  float& getRadius(){ /*Error("getRadius");*/ return radius; }
  float& getShinyRadius(){ /*Error("getShinyRadius");*/ return shiny_radius; }
  SbVec3f& getShinyCenter(){ /*Error("getShinyCenter");*/ return shiny_center; }
};

#endif /* RAYACCELCLASS */
