#include "Sphere.h"

SphereSet Sphere::defaultSphereSet;
SphereSet Sphere::defaultTextureSphereSet;
SphereSet Sphere::unitSphereSet;
SphereSet Sphere::unitTextureSphereSet;
Material  Sphere::defaultMat;
SbVec3f   Sphere::origin(0,0,0);

bool Sphere::intersectRay(Ray &r, double t0, double t1){
  
  if (r.ignorePrim == this)
    t0 = 1e-2; // expand tolerance
  r.t = -1;
  r.prim = NULL;

//     SbVec3f pp = pos()-r.pos;   // solve t^2*d.d + 2*t*(o-p).d + (o-p).(o-p)-R^2 = 0
//     double b = pp.dot(r.dir);
//     double det = b*b - pp.dot(pp) + rad()*rad();

  // solve t^2*d.d + 2*t*(o-p).d + (o-p).(o-p)-R^2 = 0    
  SbVec3f p = pos();
  double pp[3];
  pp[0] = p[0]-r.pos[0];
  pp[1] = p[1]-r.pos[1];
  pp[2] = p[2]-r.pos[2];
  double b =
    pp[0]*r.dir[0] +
    pp[1]*r.dir[1] +
    pp[2]*r.dir[2];
  double det =
    b*b -
    (pp[0]*pp[0] +
     pp[1]*pp[1] +
     pp[2]*pp[2]) +
    rad()*rad();
    
  if (det<0) return false;
  det = sqrt(det);
  double t = b-det;
  if (t>=t0 && t<t1){
    r.t = t;
    r.prim = this;
    r.bc = r.pos+r.dir*r.t;
    return true;
  }
  t = b+det;
  if (t>=t0 && t<t1) {
    r.t = t;
    r.prim = this;
    r.bc = r.pos+r.dir*r.t;
    return true;
  }
    
  return false;
}

SbColor Sphere::getIntDiffColor(const SbVec3f &bc) const {
  const SbColor &color = col();
  
  const Texture *text = texture();
  if (text && text->bumpMap !=2){
    float s, t;
    getSTfromBC(bc, s, t);
    return text->getColor(s, t, color);
  }

  return color;
}

SbVec3f Sphere::getIntNormal(const SbVec3f &bc) const {
  SbVec3f n = getNormal(bc);

  // do bump mapping?
  const Texture *text = texture();
  if (! (text && text->bumpMap))
    return n;

  // get geometric derivs if we don't have them already
  PrimDerivs myPrimDerivs;
  PrimDerivs *pDer;
//  if (!pDer){
  if (!getDerivs(bc, myPrimDerivs))
    return n;
  pDer = &myPrimDerivs;
//  }

  float s, t;
  getSTfromBC(bc, s, t);
  SbVec3f dh = text->getDeriv(s,t); // returns dh/du, dh/dv, h
  float dhdu = dh[0];        // either of these might be zero
  float dhdv = dh[1];
  float h0 = dh[2];
    
  SbVec3f dppdu = pDer->dPdU + text->bumpScale*dhdu*n + h0*pDer->dNdU;
  SbVec3f dppdv = pDer->dPdV + text->bumpScale*dhdv*n + h0*pDer->dNdV;

  SbVec3f n2 = dppdu.cross(dppdv);
  n2.normalize();
    
  return n2;
}

bool Sphere::getDerivs(const SbVec3f &bc, PrimDerivs &pDer) const {
  SbVec3f r = bc-pos();
  pDer.dPdU = SbVec3f(0,1,0).cross(r);
  pDer.dPdU.normalize();
  pDer.dPdV = r.cross(pDer.dPdU);
  pDer.dPdV.normalize();
  pDer.dNdV = pDer.dNdU = SbVec3f(0,0,0);
  return true;
}
