#include <iostream>
#include <math.h>
#include <vector>
#include <sys/time.h>
#include <unistd.h>

#include "asm/msr.h"

// compile with:
// g++ -o cycletest cycletest.cxx -I /usr/src/linux-2.4/include

using namespace std;

// Compute the delay between t1 and t2 in milliseconds
inline long delay (struct timeval t1, struct timeval t2){
  long d;
    
  d = (t2.tv_sec - t1.tv_sec) * 1000;
  d+= ((t2.tv_usec - t1.tv_usec + 500) / 1000);
  return(d);
}

int main(int argc, char *argv[]){


  struct timeval start, end;
  int sum=0, tmp;

  unsigned long ini, stp;
  unsigned long long inil, stpl;

  rdtscl(ini);
  rdtscl(stp);
  long overhead = stp-ini;

  cerr << "argc=" << argc << "\n";

  {
  cerr << "Next power of 2 loop: ";
  unsigned int x, y, i;
  int input = 1025;
  if (argc>1) input = atoi(argv[1]);
  x = input;
  rdtscll(inil);
  i=0; y=x;
  while (y=y>>1) i++;
  int r = 1<<i;
  if (r!=x)
    r = r<<1;
  rdtscll(stpl);
  cerr << r << " ";
  printf("cycles lapse: %lli\n", stpl - inil - (unsigned long long) overhead);

  const float log2 = log(2.0);
  cerr << "Next power of 2 float: ";
  x = input;
  rdtscll(inil);
  r = 1<<(int)ceilf((float)log((float)x)/log2);
  rdtscll(stpl);
  cerr << r << " ";
  printf("cycles lapse: %lli\n", stpl - inil - (unsigned long long) overhead);
  }



  

  cerr << "Divide by constant test.\n";
  const int N=100;
  float nums[N];
  for (int i=0; i<N; i++)
     nums[i] = drand48();
  rdtscl(ini);
  for (int i=0; i<N; i++)
     nums[i]/=M_PI;
     //nums[i]*=1.0/M_PI;
  rdtscl(stp);
  printf("time lapse: %li\n", stp - ini - overhead);
  for (int i=0; i<N; i++)
     sum += (int)(nums[i]*100.0);

  int ms;

  gettimeofday(&start, NULL); // start delay measure
  rdtscll(inil);
  sleep(1);
  rdtscll(stpl);
  gettimeofday(&end, NULL);   // end delay measure
  printf("cycles lapse: %lli\n", stpl - inil - (unsigned long long) overhead);
  ms = delay(start,end);
  cerr << "Sleep(1) Took " << ms << "ms.\n";

  
  rdtscl(ini);
  tmp=argc+argc+argc+argc+argc+argc+argc+argc+argc+argc+argc;
  rdtscl(stp);
  printf("10 adds, time lapse: %li\n", stp - ini - overhead);

  sum += tmp;
   
  rdtscl(ini);
  tmp=argc*argc*argc*argc*argc*argc*argc*argc*argc*argc*argc;
  rdtscl(stp);
  printf("10 muls, time lapse: %li\n", stp - ini - overhead);

  sum += tmp;

  rdtscl(ini);
  double d = argc;
  double dtmp= d*d*d*d*d*d*d*d*d*d*d;
  rdtscl(stp);
  printf("10 double precision multiplies, time lapse: %li\n", stp - ini - overhead);

  rdtscl(ini);
  float f = argc;
  float ftmp= f*f*f*f*f*f*f*f*f*f*f;
  rdtscl(stp);
  printf("10 single precision multiplies, time lapse: %li\n", stp - ini - overhead);
   
  rdtscl(ini);
  tmp=argc/argc/argc/argc/argc/argc/argc/argc/argc/argc/argc;
  rdtscl(stp);
  printf("10 divs, time lapse: %li\n", stp - ini - overhead);

  sum += tmp;

  float tmpf = pow(2.0, 30.0);
  rdtscl(ini);
  tmp=(int)sqrt(sqrt(sqrt(sqrt(sqrt(sqrt(sqrt(sqrt(sqrt(sqrt(tmpf*(float)argc))))))))));
  rdtscl(stp);
  printf("10 sqrts, time lapse: %li\n", stp - ini - overhead);
  sum += tmp;

  rdtscl(ini);
  for (int i=0; i<N; i++)
    nums[i] = sqrt(nums[i]);
  rdtscl(stp);
  printf("%d sqrts, time lapse: %li\n", N, stp - ini - overhead);
  for (int i=0; i<N; i++)
     sum += (int)(nums[i]*100.0);
  
  return(sum);
  
  srand48(time(NULL));
  double num = 1.0e8;
  double x = 1.0;
//   double y = drand48()/(num/10.0)+(1.0-10.0/num);
//   cerr << "y=" << y << "\n";
  const double y = 2.0;
  const double yinv = 1.0/y;
  
  for (int i=0; i<(int)num; i++){

//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
//     x = x/y;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;
    x = x*yinv;

//    cerr << x << "\n";
  }

  gettimeofday(&end, NULL);   // end delay measure
  ms = delay(start,end);
  cerr << "Took " << ms << "ms.\n";
  cerr << "x = " << x << "\n";
  return(1);
    
   cerr << "vector size=" << sizeof(vector <int *>) << "\n";
   return(0);
   
   //for (int i=1; i<1000000000; i++){ }
   cerr << "beep\a\n";
   cerr << "(int)-4.5 = " << (int)-4.5 << endl;
   cerr << "e(0.0001) = " << exp(0.0001) << endl;
   cerr << "floor -1.5 = " << -0.5-floor(-0.5) << endl;
   cerr << (!1 && 0) << endl;
   cerr << (!(1 && 0)) << endl;
   cerr << "log10(3.18e-7) = " << log10(3.18e-7) << endl;
   cerr << "log10(0.0) = " << log10(0.0) << endl;
   return(0);
}
